#!/usr/bin/env python
#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for `mlonmcu` package."""


import unittest

# from mlonmcu import mlonmcu


class TestMlonmcu(unittest.TestCase):
    """Tests for `mlonmcu` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_something(self):
        """Test something."""



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import os
from enum import Enum

from mlonmcu.utils import filter_none
from mlonmcu.target.target import Target
from mlonmcu.target.metrics import Metrics

from mlonmcu.feature.features import SUPPORTED_TVM_BACKENDS
from mlonmcu.target.riscv.util import sort_extensions_canonical, join_extensions

from mlonmcu.logging import get_logger

logger = get_logger()


class Esp32C3PerfCount(Enum):
    CYCLE = 1
    INST = 2
    LD_HAZARD = 4
    JMP_HAZARD = 8
    IDLE = 16
    LOAD = 32
    STORE = 64
    JMP_UNCOND = 128
    BRANCH = 256
    BRANCH_TAKEN = 512
    INST_COMP = 1024


ESPIDF_PLATFORM_TARGET_REGISTRY = {}


def register_espidf_platform_target(target_name, t, override=False):
    global ESPIDF_PLATFORM_TARGET_REGISTRY

    if target_name in ESPIDF_PLATFORM_TARGET_REGISTRY and not override:
        raise RuntimeError(f"ESP-IDF platform target {target_name} is already registered")
    ESPIDF_PLATFORM_TARGET_REGISTRY[target_name] = t


def get_espidf_platform_targets():
    return ESPIDF_PLATFORM_TARGET_REGISTRY


class Esp32C3Target(Target):
    DEFAULTS = {
        **Target.DEFAULTS,
        "xlen": 32,
        "extensions": ["i", "m", "c"],
        "fpu": "none",
        "arch": None,
        "abi": None,
        "attr": "",
        "count": Esp32C3PerfCount.CYCLE,
    }

    @property
    def xlen(self):
        return int(self.config["xlen"])

    @property
    def extensions(self):
        exts = self.config.get("extensions", []).copy()
        if not isinstance(exts, list):
            exts = exts.split(",")
        if "g" not in exts:
            required = []
            if self.fpu == "double":
                required.append("d")
                required.append("f")
            if self.fpu == "single":
                required.append("f")
            for ext in required:
                if ext not in exts:
                    exts.append(ext)
        return exts

    @property
    def arch(self):
        temp = self.config["arch"]  # TODO: allow underscores and versions
        if temp:
            return temp
        else:
            exts_str = join_extensions(sort_extensions_canonical(self.extensions, lower=True))
            return f"rv{self.xlen}{exts_str}"

    @property
    def abi(self):
        temp = self.config["abi"]
        if temp:
            return temp
        else:
            if self.xlen == 32:
                temp = "ilp32"
            elif self.xlen == 64:
                temp = "lp64"
            else:
                raise RuntimeError(f"Invalid xlen: {self.xlen}")
            if "d" in self.extensions or "g" in self.extensions:
                temp += "d"
            elif "f" in self.extensions:
                temp += "f"
            return temp

    @property
    def attr(self):
        attrs = str(self.config["attr"]).split(",")
        if len(attrs) == 1 and len(attrs[0]) == 0:
            attrs = []
        for ext in sort_extensions_canonical(self.extensions, lower=True, unpack=True):
            attrs.append(f"+{ext}")
        attrs = list(set(attrs))
        return ",".join(attrs)

    @property
    def fpu(self):
        value = self.config["fpu"]
        if value is None or not value:
            value = "none"
        assert value in ["none", "single", "double"]
        return value

    @property
    def count(self):
        value = int(self.config["count"])
        return value

    @property
    def has_fpu(self):
        return self.fpu != "none"

    def get_platform_defs(self, platform):
        ret = super().get_platform_defs(platform)
        ret["ESP32C3_PERF_COUNT"] = self.count
        return ret

    def get_target_system(self):
        return "esp32c3"

    def get_arch(self):
        return "riscv"

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = {}
        if backend in SUPPORTED_TVM_BACKENDS:
            ret.update(
                {
                    "target_device": "riscv_cpu",
                    "target_march": self.arch,
                    "target_model": "esp32c3_devkit",
                    "target_mtriple": "riscv32-esp-elf",
                    "target_mabi": self.abi,
                    "target_mattr": self.attr,
                    "target_mcpu": "esp32c3",
                }
            )
            if optimized_schedules:
                ret.update(
                    {
                        "target_device": "riscv_cpu",
                    }
                )
        return ret

    def add_backend_config(self, backend, config, optimized_layouts=False, optimized_schedules=False):
        new = filter_none(
            self.get_backend_config(
                backend, optimized_layouts=optimized_layouts, optimized_schedules=optimized_schedules
            )
        )

        # only allow overwriting non-none values
        # to support accepting user-vars
        new = {key: value for key, value in new.items() if config.get(key, None) is None}
        config.update(new)


def create_espidf_platform_target(name, platform, base=Target):
    class EspIdfPlatformTarget(base):
        DEFAULTS = {
            **base.DEFAULTS,
            "timeout_sec": 0,  # disabled
            "port": None,
            "baud": None,
        }

        def __init__(self, features=None, config=None):
            super().__init__(name=name, features=features, config=config)
            self.platform = platform

        @property
        def timeout_sec(self):
            return int(self.config["timeout_sec"])

        @property
        def port(self):
            return self.config["port"]

        @property
        def baud(self):
            return self.config["baud"]

        def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
            """Use target to execute a executable with given arguments"""
            if len(args) > 0:
                raise RuntimeError("Program arguments are not supported for real hardware devices")

            assert self.platform is not None, "ESP32 targets need a platform to execute programs"

            if self.timeout_sec > 0:
                raise NotImplementedError

            # ESP-IDF actually wants a project directory, but we only get the elf now. As a workaround we
            # assume the elf is right in the build directory inside the project directory

            ret = self.platform.run(program, self)
            return ret

        def parse_stdout(self, out):
            cpu_cycles = re.search(r"Total Cycles: (.*)", out)
            if not cpu_cycles:
                logger.warning("unexpected script output (cycles)")
                cycles = None
            else:
                cycles = int(float(cpu_cycles.group(1)))
            cpu_time_us = re.search(r"Total Time: (.*) us", out)
            if not cpu_time_us:
                logger.warning("unexpected script output (time_us)")
                time_us = None
            else:
                time_us = int(float(cpu_time_us.group(1)))
            return cycles, time_us

        def get_metrics(self, elf, directory, handle_exit=None):
            if self.print_outputs:
                out = self.exec(elf, cwd=directory, live=True, handle_exit=handle_exit)
            else:
                out = self.exec(
                    elf, cwd=directory, live=False, print_func=lambda *args, **kwargs: None, handle_exit=handle_exit
                )
            cycles, time_us = self.parse_stdout(out)

            metrics = Metrics()
            metrics.add("Cycles", cycles)
            time_s = time_us / 1e6 if time_us is not None else time_us
            metrics.add("Runtime [s]", time_s)

            return metrics, out, []

        def get_arch(self):
            return "unkwown"

    return EspIdfPlatformTarget



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import _validate_gcc, get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_corstone300(context: MlonMcuContext, params=None):
    return context.environment.has_target("corstone300") and _validate_gcc(context, params=params)


@Tasks.provides(["corstone300.exe"])
@Tasks.validate(_validate_corstone300)
@Tasks.register(category=TaskType.TARGET)
def install_corstone300(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Download and install corstone300 FVP from ARM."""
    fvpName = utils.makeDirName("corstone300")
    fvpInstallDir = context.environment.paths["deps"].path / "install" / fvpName
    fvpSubDir = fvpInstallDir / "fvp"
    fvpExe = fvpSubDir / "models" / "Linux64_GCC-6.4" / "FVP_Corstone_SSE-300_Ethos-U55"
    user_vars = context.environment.vars
    if "corstone300.exe" in user_vars:  # TODO: also check command line flags?
        # fvpExe = user_vars["corstone300.exe"]
        return False
    else:
        if not fvpExe.is_file():
            fvpUrl = "https://developer.arm.com/-/media/Arm%20Developer%20Community/Downloads/OSS/FVP/Corstone-300/"
            fvpFileName = "FVP_Corstone_SSE-300_11.16_26"
            fvpArchive = fvpFileName + ".tgz"
            utils.download_and_extract(fvpUrl, fvpArchive, fvpInstallDir, progress=verbose)
            fvpScript = fvpInstallDir / "FVP_Corstone_SSE-300.sh"
            utils.exec_getout(
                fvpScript,
                "--i-agree-to-the-contained-eula",
                "--no-interactive",
                "-d",
                fvpSubDir,
            )
    context.cache["corstone300.exe"] = fvpExe
    context.export_paths.add(fvpExe.parent)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import os
import pkg_resources
from pathlib import Path
import multiprocessing

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()
Tasks = get_task_factory()

#############
# pulp-gcc  #
#############


def _validate_pulp(context: MlonMcuContext, params=None):
    return context.environment.has_target("gvsoc_pulp") or context.environment.has_target("gvsoc_pulpissimo")


def _validate_pulp_gcc(context: MlonMcuContext, params=None):
    return context.environment.has_toolchain("gcc") and _validate_pulp(context, params)


@Tasks.provides(["pulp_gcc.install_dir", "pulp_gcc.name"])
@Tasks.validate(_validate_pulp_gcc)
@Tasks.register(category=TaskType.TOOLCHAIN)
def install_pulp_gcc(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Download and install the PULP (RISCV) GCC toolchain."""
    if not params:
        params = {}
    user_vars = context.environment.vars
    from_source = user_vars.get("pulp_gcc.from_source", False)
    flags = utils.makeFlags()
    pulpGccName = utils.makeDirName("pulp_gcc", flags=flags)
    pulpGccInstallDir = context.environment.paths["deps"].path / "install" / pulpGccName
    if "pulp_gcc.install_dir" in user_vars:  # TODO: also check command line flags?
        # This would overwrite the cache.ini entry which is NOT wanted! -> return false but populate gcc_name?
        pulpGccInstallDir = user_vars["pulp_gcc.install_dir"]
    else:
        if rebuild or not utils.is_populated(pulpGccInstallDir):
            if from_source:
                pulpGccSrcDir = context.environment.paths["deps"].path / "src" / pulpGccName
                pulpGccBuildDir = context.environment.paths["deps"].path / "build" / pulpGccName
                utils.mkdirs(pulpGccBuildDir)
                if rebuild or not utils.is_populated(pulpGccSrcDir):
                    pulpGccRepo = context.environment.repos["pulp_gcc"]
                    utils.clone_wrapper(pulpGccRepo, pulpGccSrcDir, refresh=rebuild, recursive=True)
                env = os.environ.copy()
                env["PATH"] = str(pulpGccInstallDir) + ":" + env["PATH"]
                if rebuild or not utils.is_populated(pulpGccBuildDir):
                    pulpGccArgs = []
                    pulpGccArgs.append("--prefix=" + str(pulpGccInstallDir))
                    pulpGccArgs.append("--with-arch=rv32imc")
                    pulpGccArgs.append("--with-cmodel=medlow")
                    pulpGccArgs.append("--enable-multilib")
                    utils.exec_getout(
                        str(pulpGccSrcDir / "configure"),
                        *pulpGccArgs,
                        cwd=pulpGccBuildDir,
                        env=env,
                        live=verbose,
                    )
                utils.make(
                    cwd=pulpGccBuildDir,
                    env=env,
                    live=verbose,
                    threads=threads,
                )
            else:
                pulpGccUrl = user_vars.get("pulp_gcc.dl_url", None)
                assert pulpGccUrl, "pulp_gcc.dl_url undefined and environment and pulp_gcc.from_source=0"
                pulpGccBaseUrl, pulpGccArchive = pulpGccUrl.rsplit("/", 1)
                utils.download_and_extract(pulpGccBaseUrl, pulpGccArchive, pulpGccInstallDir)
    if "pulp_gcc.name" in user_vars:
        pulpGccName = user_vars["pulp_gcc.name"]
    else:
        pulpGccName = "riscv32-unknown-elf"
    context.cache["pulp_gcc.install_dir", flags] = pulpGccInstallDir
    context.cache["pulp_gcc.name", flags] = pulpGccName
    # context.cache["pulp_gcc.build_dir", flags] = pulpGccBuildDir


@Tasks.provides(["pulp_freertos.src_dir", "pulp_freertos.support_dir", "pulp_freertos.config_dir"])
@Tasks.validate(_validate_pulp)
@Tasks.register(category=TaskType.TARGET)
def clone_pulp_freertos(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the pulp-freertos repository."""
    pulpRtosName = utils.makeDirName("pulp_freertos")
    pulpRtosSrcDir = context.environment.paths["deps"].path / "src" / pulpRtosName
    pulpRtosSupportDir = pulpRtosSrcDir / "support"
    pulpConfigsDir = pulpRtosSupportDir / "pulp-configs" / "configs"
    if (
        rebuild
        or not utils.is_populated(pulpRtosSrcDir)
        or not utils.is_populated(pulpRtosSupportDir)
        or not utils.is_populated(pulpConfigsDir)
    ):
        pulpRtosRepo = context.environment.repos["pulp_freertos"]
        utils.clone_wrapper(pulpRtosRepo, pulpRtosSrcDir, refresh=rebuild)
        user_vars = context.environment.vars
        experimental_install = user_vars.get("pulp_freertos.experimental_install", False)
        if experimental_install:
            patchFile = Path(
                pkg_resources.resource_filename(
                    "mlonmcu", os.path.join("..", "resources", "patches", "pulp_freertos_support.patch")
                )
            )
            if patchFile.is_file():
                utils.patch(patchFile, cwd=pulpRtosSrcDir)
    context.cache["pulp_freertos.src_dir"] = pulpRtosSrcDir
    context.cache["pulp_freertos.support_dir"] = pulpRtosSupportDir
    context.cache["pulp_freertos.config_dir"] = pulpConfigsDir


@Tasks.needs(["pulp_freertos.src_dir", "pulp_freertos.support_dir", "pulp_freertos.config_dir"])
@Tasks.provides(["pulp_freertos.install_dir", "pulp_freertos.pythonpath", "gvsoc.exe"])
@Tasks.validate(_validate_pulp)
@Tasks.register(category=TaskType.TARGET)
def install_gvsoc(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Setup the pulp-freertos build."""
    pulpRtosName = utils.makeDirName("pulp_freertos")
    user_vars = context.environment.vars
    experimental_install = user_vars.get("pulp_freertos.experimental_install", False)
    pulpRtosSupportDir = context.cache["pulp_freertos.support_dir"]
    gvsocExe = pulpRtosSupportDir / "egvsoc.sh"
    pulpConfigDir = context.cache["pulp_freertos.config_dir"]
    if experimental_install:
        pulpRtosInstallDir = context.environment.paths["deps"].path / "install" / pulpRtosName
    else:
        pulpRtosInstallDir = pulpRtosSupportDir / "install"
    pulpPythonPath = pulpRtosInstallDir / "python"
    if rebuild or not utils.is_populated(pulpRtosInstallDir) or not utils.is_populated(pulpPythonPath):
        utils.mkdirs(pulpRtosInstallDir)
        env = os.environ.copy()
        env.update(
            {
                "PULP_CURRENT_CONFIG": "pulpissimo@config_file=chips/pulpissimo/pulpissimo.json",
                "PULP_CONFIGS_PATH": pulpConfigDir,
                "PYTHONPATH": pulpRtosInstallDir / "python",
                "INSTALL_DIR": pulpRtosInstallDir,
                "ARCHI_DIR": pulpRtosSupportDir / "archi" / "include",
                "SUPPORT_ROOT": pulpRtosSupportDir,
            }
        )
        if experimental_install:
            env.update(
                {
                    "SUPPORT_INSTALL_DIR": pulpRtosInstallDir,
                }
            )
        makeArgs = ["-f", pulpRtosSupportDir / "support.mk", "gvsoc"]
        utils.make(
            *makeArgs,
            cwd=pulpRtosInstallDir,
            env=env,
            live=verbose,
            threads=1,  # Script returns 2 exit code if parallel
        )
    context.cache["pulp_freertos.install_dir"] = pulpRtosInstallDir
    context.cache["pulp_freertos.pythonpath"] = pulpPythonPath
    context.cache["gvsoc.exe"] = gvsocExe
    context.export_paths.add(gvsocExe.parent)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for the load stage."""

from mlonmcu.cli.common import (
    add_common_options,
    add_context_options,
    add_model_options,
    add_flow_options,
    kickoff_runs,
)

from .helper.parse import extract_config_and_feature_names, extract_frontend_names, extract_postprocess_names
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.models import SUPPORTED_FRONTENDS
from mlonmcu.models.lookup import apply_modelgroups
from mlonmcu.session.run import RunStage


def add_load_options(parser):
    add_common_options(parser)
    add_context_options(parser)
    add_flow_options(parser)
    add_model_options(parser)
    load_parser = parser.add_argument_group("load options")
    load_parser.add_argument(
        "--frontend",
        type=str,
        metavar="FRONTEND",
        choices=SUPPORTED_FRONTENDS.keys(),
        default=None,
        nargs=1,
        help="Explicitly choose the frontends to use (choices: %(choices)s)",
    )


def get_parser(subparsers):
    """ "Define and return a subparser for the load subcommand."""
    parser = subparsers.add_parser("load", description="Load model using the ML on MCU flow.")
    parser.set_defaults(flow_func=handle)
    add_load_options(parser)
    return parser


def _handle(args, context):
    config = context.environment.vars
    new_config, features, gen_config, gen_features = extract_config_and_feature_names(args, context=context)
    config.update(new_config)
    frontends = extract_frontend_names(args, context=context)
    postprocesses = extract_postprocess_names(args, context=context)
    session = context.get_session(label=args.label, resume=args.resume, config=config)
    models = apply_modelgroups(args.models, context=context)
    for model in models:
        for f in gen_features:
            for c in gen_config:
                all_config = {**config, **c}
                run = session.create_run(config=all_config)
                all_features = list(set(features + f))
                run.add_features_by_name(all_features, context=context)  # TODO do this before load.py?
                run.add_frontends_by_name(frontends, context=context)
                run.add_model_by_name(model, context=context)
                run.add_postprocesses_by_name(postprocesses, context=context)  # TODO do this before load.py?


def handle(args, ctx=None):
    if ctx:
        _handle(args, ctx)
    else:
        with MlonMcuContext(path=args.home, deps_lock="read") as context:
            _handle(args, context)
            kickoff_runs(args, RunStage.LOAD, context)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU Spike Target definitions"""

from mlonmcu.logging import get_logger
from mlonmcu.config import str2bool
from mlonmcu.utils import is_power_of_two
from mlonmcu.feature.features import SUPPORTED_TVM_BACKENDS
from .riscv import RISCVTarget
from .util import update_extensions

logger = get_logger()


class RVVTarget(RISCVTarget):
    """TODO"""

    FEATURES = RISCVTarget.FEATURES | {"vext"}

    DEFAULTS = {
        **RISCVTarget.DEFAULTS,
        "enable_vext": False,
        "vext_spec": 1.0,
        "embedded_vext": False,
        "vlen": 128,  # vectorization=off
        "elen": 64,
    }
    REQUIRED = RISCVTarget.REQUIRED

    def __init__(self, name, features=None, config=None):
        super().__init__(name, features=features, config=config)
        self.supported_vext_spec_min = 1.0
        self.supported_vext_spec_max = 1.0

    @property
    def enable_vext(self):
        value = self.config["enable_vext"]
        return str2bool(value)

    @property
    def vlen(self):
        value = int(self.config["vlen"])
        assert value == 0 or is_power_of_two(value), "VLEN needs to be a power of 2."
        assert value == 0 or value >= 32, "VLEN < 32 not allowed"
        if 0 < value < 128:
            assert self.embedded_vext, "VLEN < 128 imples embedded_vext=false"
        return value

    @property
    def elen(self):
        value = int(self.config["elen"])
        assert value in [32, 64]
        if value == 32:
            assert self.embedded_vext, "ELEN=32 imples embedded_vext=true"
        return value

    @property
    def vext_spec(self):
        return float(self.config["vext_spec"])

    @property
    def embedded_vext(self):
        value = self.config["embedded_vext"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def extensions(self):
        exts = super().extensions
        assert (
            self.supported_vext_spec_min <= self.vext_spec <= self.supported_vext_spec_max
        ), f"V-Extension spec {self.vext_spec} not supported by {self.name} target"
        return update_extensions(
            exts,
            vext=self.enable_vext,
            elen=self.elen,
            embedded_vext=self.embedded_vext,
            vlen=self.vlen,
            fpu=self.fpu,
        )

    @property
    def attrs(self):
        attrs = super().attrs
        if self.enable_vext and f"+zvl{self.vlen}b" not in attrs:
            attrs.append(f"+zvl{self.vlen}b")
        return attrs

    def get_platform_defs(self, platform):
        ret = super().get_platform_defs(platform)
        if self.enable_vext:
            major, minor = str(self.vext_spec).split(".")[:2]
            ret["RISCV_RVV_MAJOR"] = major
            ret["RISCV_RVV_MINOR"] = minor
            ret["RISCV_RVV_VLEN"] = self.vlen
        return ret

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = super().get_backend_config(
            backend, optimized_layouts=optimized_layouts, optimized_schedules=optimized_schedules
        )
        if backend in SUPPORTED_TVM_BACKENDS:
            model = ret["target_model"]
            if self.enable_vext:
                if "zvl" not in model:
                    model = f"{model}-zvl{self.vlen}b"
            ret["target_model"] = model
        return ret



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from mlonmcu.models.lookup import print_summary
from .frontend import (
    PBFrontend,
    TfLiteFrontend,
    PackedFrontend,
    ONNXFrontend,
    RelayFrontend,
    PaddleFrontend,
    ExampleFrontend,
    EmbenchFrontend,
    TaclebenchFrontend,
    PolybenchFrontend,
    CoremarkFrontend,
    DhrystoneFrontend,
    MathisFrontend,
    MibenchFrontend,
    LayerGenFrontend,
)

SUPPORTED_FRONTENDS = {
    "tflite": TfLiteFrontend,
    "relay": RelayFrontend,
    "packed": PackedFrontend,
    "onnx": ONNXFrontend,
    "pb": PBFrontend,
    "paddle": PaddleFrontend,
    "example": ExampleFrontend,
    "embench": EmbenchFrontend,
    "taclebench": TaclebenchFrontend,
    "coremark": CoremarkFrontend,
    "dhrystone": DhrystoneFrontend,
    "polybench": PolybenchFrontend,
    "mathis": MathisFrontend,
    "mibench": MibenchFrontend,
    "layergen": LayerGenFrontend,
}  # TODO: use registry instead

__all__ = [
    "print_summary",
    "TfLiteFrontend",
    "PackedFrontend",
    "ONNXFrontend",
    "PBFrontend",
    "LayerGenFrontend",
    "SUPPORTED_FRONTENDS",
]



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU OVPSim Target definitions"""

import os
import re
from pathlib import Path

from mlonmcu.logging import get_logger
from mlonmcu.config import str2bool
from mlonmcu.feature.features import SUPPORTED_TVM_BACKENDS
from mlonmcu.setup.utils import execute
from mlonmcu.target.common import cli
from mlonmcu.target.metrics import Metrics
from .riscv import sort_extensions_canonical
from .util import update_extensions
from .riscv_pext_target import RVPTarget
from .riscv_vext_target import RVVTarget

logger = get_logger()


MAX_P_SPEC = 0.96


def replace_unsupported(exts):
    ret = []
    for ext in exts:
        if "ZVE" in ext or "ZVL" in ext:
            ret.append("V")
        if "zpn" in ext.lower() or "zpsfoperand" in ext.lower():
            pass
        elif ext == "P":
            ret.append(ext)
            ret.append("B")
        else:
            ret.append(ext)
    return ret


class OVPSimTarget(RVPTarget, RVVTarget):
    """Target using an ARM FVP (fixed virtual platform) based on a Cortex M55 with EthosU support"""

    FEATURES = RVPTarget.FEATURES | RVVTarget.FEATURES | {"gdbserver", "log_instrs", "trace"}

    DEFAULTS = {
        **RVPTarget.DEFAULTS,
        **RVVTarget.DEFAULTS,
        "bitmanip_spec": 0.94,
        # TODO: add bext feature
        "variant": None,
        "end_to_end_cycles": True,
        "gdbserver_enable": False,
        "gdbserver_attach": False,
        "gdbserver_port": 2222,
    }
    REQUIRED = RVPTarget.REQUIRED | RVVTarget.REQUIRED | {"ovpsim.exe"}

    def __init__(self, name="ovpsim", features=None, config=None):
        super().__init__(name, features=features, config=config)

    @property
    def ovpsim_exe(self):
        return Path(self.config["ovpsim.exe"])

    @property
    def variant(self):
        temp = self.config["variant"]
        if temp:
            return temp
        else:
            return f"RVB{self.xlen}I"

    @property
    def extensions(self):
        # exts = RVPTarget.extensions(self) + RVVTarget.extensions(self)
        exts = super().extensions
        return update_extensions(
            exts,
        )

    @property
    def end_to_end_cycles(self):
        value = self.config["end_to_end_cycles"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def gdbserver_enable(self):
        value = self.config["gdbserver_enable"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def gdbserver_attach(self):
        value = self.config["gdbserver_attach"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def gdbserver_port(self):
        return int(self.config["gdbserver_port"])

    def get_default_ovpsim_args(self):
        extensions_before = sort_extensions_canonical(self.extensions, lower=False, unpack=True)
        extensions_after = replace_unsupported(extensions_before)
        extensions_str = "".join(sort_extensions_canonical(extensions_after))
        args = [
            "--variant",
            self.variant,
            "--override",
            f"riscvOVPsim/cpu/add_Extensions={extensions_str}",
            "--override",
            "riscvOVPsim/cpu/unaligned=T",
            "--override",
            "riscvOVPsim/cpu/pk/reportExitErrors=T",
            "--finishonopcode",
            "0",
        ]
        if self.enable_pext:
            args.extend(
                [
                    "--override",
                    f"riscvOVPsim/cpu/dsp_version={self.pext_spec}",
                    "--override",
                    f"riscvOVPsim/cpu/bitmanip_version={self.bitmanip_spec}",
                ]
            )
        if self.enable_vext:
            assert self.has_fpu, "V-Extension requires enabled FPU"
            args.extend(
                [
                    "--override",
                    "riscvOVPsim/cpu/vector_version=1.0-draft-20210130",  # TODO: use vext_spec
                    "--override",
                    f"riscvOVPsim/cpu/VLEN={self.vlen}",
                    "--override",
                    f"riscvOVPsim/cpu/ELEN={self.elen}",
                ]
            )
            args.extend(["--override", f"riscvOVPsim/cpu/mstatus_VS={int(self.enable_vext)}"])
        if self.has_fpu:
            assert "f" in self.extensions or "g" in self.extensions
        args.extend(["--override", f"riscvOVPsim/cpu/mstatus_FS={int(self.has_fpu)}"])
        if self.gdbserver_enable:
            # args.append("--trace")
            args.extend(["--port", str(self.gdbserver_port)])
            if self.gdbserver_attach:
                args.append("--gdbconsole")
        return args

    def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
        """Use target to execute a executable with given arguments"""
        ovpsim_args = []

        ovpsim_args.extend(["--program", str(program)])
        ovpsim_args.extend(self.get_default_ovpsim_args())

        if len(self.extra_args) > 0:
            if isinstance(self.extra_args, str):
                extra_args = self.extra_args.split(" ")
            else:
                extra_args = self.extra_args
            ovpsim_args.extend(extra_args)  # I rename args to extra_args because otherwise it overwrites *args

        if self.timeout_sec > 0:
            raise NotImplementedError

        ret = execute(
            self.ovpsim_exe.resolve(),
            *ovpsim_args,
            *args,  # Does this work?
            cwd=cwd,
            **kwargs,
        )
        return ret

    def parse_stdout(self, out):
        # cpi = 1
        if self.end_to_end_cycles:
            cpu_cycles = re.search(r".*  Simulated instructions:(.*)", out)
        else:
            cpu_cycles = re.search(r".* Total Cycles: (.*)", out)
        if not cpu_cycles:
            raise RuntimeError("unexpected script output (cycles)")
            cycles = None
        else:
            cycles = int(cpu_cycles.group(1).replace(",", ""))
        mips = None  # TODO: parse mips?
        mips_match = re.search(r".*  Simulated MIPS:(.*)", out)
        if mips_match:
            mips_str = float(mips_match.group(1))
            if "run too short for meaningful result" not in mips:
                mips = float(mips_str)
        return cycles, mips

    def get_metrics(self, elf, directory, *args, handle_exit=None):
        out = ""
        if self.print_outputs:
            out += self.exec(elf, *args, cwd=directory, live=True, handle_exit=handle_exit)
        else:
            out += self.exec(
                elf, *args, cwd=directory, live=False, print_func=lambda *args, **kwargs: None, handle_exit=handle_exit
            )
        cycles, mips = self.parse_stdout(out)

        metrics = Metrics()
        metrics.add("Cycles", cycles)
        if mips:
            metrics.add("MIPS", mips, optional=True)

        return metrics, out, []

    def get_platform_defs(self, platform):
        ret = {}
        ret.update(RVPTarget.get_platform_defs(self, platform))
        ret.update(RVVTarget.get_platform_defs(self, platform))
        return ret

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = super().get_backend_config(
            backend, optimized_layouts=optimized_layouts, optimized_schedules=optimized_schedules
        )
        if backend in SUPPORTED_TVM_BACKENDS:
            if optimized_layouts:
                if self.enable_pext or self.enable_vext:
                    ret.update(
                        {
                            "desired_layout": "NHWC:HWOI",
                        }
                    )
        return ret


if __name__ == "__main__":
    cli(target=OVPSimTarget)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Strategies for the vanilla_accelerator accelerator"""

# Example how to integrate a custom conv1d strategy:

# @relay.op.strategy.override_native_generic_func("custom_conv1d_strategy")
# def custom_conv1d_strategy(attrs, inputs, out_type, target):
#     strategy = _op.OpStrategy()
#     strategy.add_implementation(
#         wrap_compute_conv1d(custom_conv1d_compute),
#         wrap_topi_schedule(custom_conv1d_schedule),
#         name="custom_conv1d.generic",
#     return strategy
#

# For further details see:
# - github.com/apache/tvm-rfcs/blob/main/rfcs/0060_UMA_Unified_Modular_Accelerator_Interface.md
# - $TVM_HOME/python/tvm/relay/op/strategy/x86.py



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from mlonmcu.logging import get_logger

from mlonmcu.flow.tvm.backend.backend import TVMBackend

from mlonmcu.flow.tvm.backend.tvmaot import TVMAOTBackend

# from mlonmcu.flow.tvm.backend.tvmaotplus import TVMAOTPlusBackend
from mlonmcu.flow.tvm.backend.tvmllvm import TVMLLVMBackend
from mlonmcu.flow.tvm.backend.tvmrt import TVMRTBackend

logger = get_logger()

MICROTVM_PLATFORM_BACKEND_REGISTRY = {}


def register_microtvm_platform_backend(backend_name, b, override=False):
    global MICROTVM_PLATFORM_BACKEND_REGISTRY

    if backend_name in MICROTVM_PLATFORM_BACKEND_REGISTRY and not override:
        raise RuntimeError(f"TVM platform backend  {backend_name} is already registered")
    MICROTVM_PLATFORM_BACKEND_REGISTRY[backend_name] = b


def get_microtvm_platform_backends():
    return MICROTVM_PLATFORM_BACKEND_REGISTRY


register_microtvm_platform_backend("tvmaot", TVMAOTBackend)
# register_microtvm_platform_backend("tvmaotplus", TVMAOTPlusBackend)
register_microtvm_platform_backend("tvmrt", TVMRTBackend)
register_microtvm_platform_backend("tvmllvm", TVMLLVMBackend)


def create_microtvm_platform_backend(name, platform, base=TVMBackend):
    class MicroTvmPlatformBackend(base):
        def __init__(self, features=None, config=None):
            super().__init__(runtime="crt", fmt="mlf", system_lib=True, features=features, config=config)
            self.platform = platform

    return MicroTvmPlatformBackend



import pytest

from mlonmcu.models.convert_data import convert


def test_convert_float():
    # valid
    assert convert("float", "1.0,2.0,3.0,4.0") == b"\x00\x00\x80\x3f\x00\x00\x00\x40\x00\x00\x40\x40\x00\x00\x80\x40"

    # invalid
    with pytest.raises(ValueError):
        convert("float", "\x12\x34\x56\x78")


def test_convert_hexstr():
    # valid
    assert convert("hexstr", "\\x12\\x34\\x56\\x78") == b"\x12\x34\x56\x78"


def test_convert_int8():
    # valid
    assert convert("int8", "1,2,3,4") == b"\x01\x02\x03\x04"

    # invalid
    with pytest.raises(ValueError):
        convert("int8", "\x12\x34\x56\x78")
    with pytest.raises(ValueError):
        convert("int8", "1.0,2.0,3.0,4.0")


def test_convert_invalid_mode():
    with pytest.raises(AssertionError):
        convert("invalid", "\x12\x34\x56\x78")



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Zephyr Platform"""

import re
import os
import time
import shutil
import tempfile
from pathlib import Path
import pkg_resources
from typing import Tuple


from mlonmcu.setup import utils
from mlonmcu.artifact import Artifact, ArtifactFormat
from mlonmcu.logging import get_logger
from mlonmcu.target import get_targets
from mlonmcu.target.target import Target
from mlonmcu.config import str2bool

from ..platform import CompilePlatform, TargetPlatform
from .zephyr_target import create_zephyr_platform_target

logger = get_logger()


def get_project_template(name="project2"):  # Workaround which only support tvmaot!!!
    zephyr_templates = pkg_resources.resource_listdir("mlonmcu", os.path.join("..", "resources", "platforms", "zephyr"))
    if name not in zephyr_templates:
        return None
    fname = pkg_resources.resource_filename("mlonmcu", os.path.join("..", "resources", "platforms", "zephyr", name))
    return fname


class ZephyrPlatform(CompilePlatform, TargetPlatform):
    """Zephyr Platform class."""

    FEATURES = CompilePlatform.FEATURES | TargetPlatform.FEATURES | {"benchmark"}

    DEFAULTS = {
        **CompilePlatform.DEFAULTS,
        **TargetPlatform.DEFAULTS,
        "project_template": None,
        "project_dir": None,
        "port": None,
        # "port": "/dev/ttyUSB0",
        "baud": 115200,
        "wait_for_user": True,
        "flash_only": False,
        "optimize": None,  # values: 0,1,2,3,s
    }

    REQUIRED = {"zephyr.install_dir", "zephyr.sdk_dir", "zephyr.venv_dir"}

    def __init__(self, features=None, config=None):
        super().__init__(
            "zephyr",
            features=features,
            config=config,
        )
        self.tempdir = None
        self.project_name = "app"
        self.project_dir = None

    @property
    def zephyr_install_dir(self):
        return Path(self.config["zephyr.install_dir"])

    @property
    def zephyr_sdk_dir(self):
        return Path(self.config["zephyr.sdk_dir"])

    @property
    def zephyr_venv_dir(self):
        return Path(self.config["zephyr.venv_dir"])

    @property
    def wait_for_user(self):
        value = self.config["wait_for_user"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def flash_only(self):
        # TODO: get rid of this
        value = self.config["flash_only"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def optimize(self):
        return self.config["optimize"]

    def invoke_west(self, *args, **kwargs):
        env = os.environ.copy()
        env["ZEPHYR_BASE"] = str(self.zephyr_install_dir / "zephyr")
        env["ZEPHYR_SDK_INSTALL_DIR"] = str(self.zephyr_sdk_dir)
        cmd = ". " + str(self.zephyr_venv_dir / "bin" / "activate") + " && west " + " ".join([str(arg) for arg in args])
        out = utils.exec_getout(
            cmd, shell=True, env=env, **kwargs, executable="/bin/bash"
        )  # TODO: using shell=True is insecure but right now we can not avoid it?
        return out

    def init_directory(self, path=None, context=None):
        if self.project_dir is not None:
            self.project_dir.mkdir(exist_ok=True)
            logger.debug("Project directory already initialized")
            return
        dir_name = self.name
        if path is not None:
            self.project_dir = Path(path)
        elif self.config["project_dir"] is not None:
            self.project_dir = Path(self.config["project_dir"])
        else:
            if context:
                assert "temp" in context.environment.paths
                self.project_dir = (
                    context.environment.paths["temp"].path / dir_name
                )  # TODO: Need to lock this for parallel builds
            else:
                logger.debug(
                    "Creating temporary directory because no context was available "
                    "and 'zephyr.project_dir' was not supplied"
                )
                self.tempdir = tempfile.TemporaryDirectory()
                self.project_dir = Path(self.tempdir.name) / dir_name
                logger.debug("Temporary project directory: %s", self.project_dir)
        self.project_dir.mkdir(exist_ok=True)

    def get_supported_targets(self):
        with tempfile.TemporaryDirectory() as temp:
            f = Path(temp) / "CMakeLists.txt"
            # f.touch()
            with open(f, "w") as handle:
                handle.write(
                    """
cmake_minimum_required(VERSION 3.22)
find_package(Zephyr REQUIRED HINTS $ENV{ZEPHYR_BASE})
project(ProjectName)
                """
                )
            b = Path(temp) / "build"

            # This will fail
            # Workaround used to get list of supported boards...
            def _handle(code):
                return 0

            text = self.invoke_west("build", "-d", b, "-b", "help", temp, live=False, handle_exit=_handle)
        # Warning: This will fail if a python executable is NOT available in the system. Aliasing
        # python3 to python will not work. Not sure how this would handle a system which only has python2 installed?
        target_names = re.compile(r"^  (\S+)$", re.MULTILINE).findall(text)

        return [
            f"zephyr_{name}" for name in target_names if len(name) > 0 and " " not in name
        ]  # TODO: consider dropping zephyr_ prefix

    def create_target(self, name):
        assert name in self.get_supported_targets(), f"{name} is not a valid Zephyr target"
        targets = get_targets()
        if name in targets:
            base = targets[name]
        else:
            base = Target
        return create_zephyr_platform_target(name, self, base=base)

    @property
    def project_template(self):
        return self.config["project_template"]

    @property
    def port(self):
        return self.config["port"]

    @property
    def baud(self):
        return self.config["baud"]

    @property
    def build_dir(self):
        return self.project_dir / "build"

    def close(self):
        if self.tempdir:
            self.tempdir.cleanup()

    def get_west_cmake_args(self):
        cmake_defs = {"CMAKE_BUILD_TYPE": "Debug" if self.debug else "Release"}
        return [f"-D{key}={value}" for key, value in cmake_defs.items()]

    def prepare(self, target, src):
        self.init_directory()
        template_dir = self.project_template
        if template_dir is None:
            template_dir = get_project_template()
        else:
            template_dir = Path(template_dir)
            if not template_dir.is_dir():
                template_dir = get_project_template(name=str(template_dir))
        assert template_dir is not None, f"Provided project template does not exists: {template_dir}"
        shutil.copytree(template_dir, self.project_dir, dirs_exist_ok=True)

        def write_defaults(filename):
            defs = self.definitions
            with open(filename, "w", encoding="utf-8") as f:
                f.write("CONFIG_CPLUSPLUS=y\n")
                f.write("CONFIG_NEWLIB_LIBC=y\n")
                f.write("CONFIG_REBOOT=y\n")
                f.write("CONFIG_STD_CPP20=y\n")
                f.write("CONFIG_LIB_CPLUSPLUS=y\n")
                f.write("CONFIG_TIMING_FUNCTIONS=y\n")
                # f.write("CONFIG_NEWLIB_LIBC_FLOAT_PRINTF=y\n")
                # f.write("CONFIG_CBPRINTF_FP_SUPPORT=y\n")
                if self.debug:
                    f.write("CONFIG_DEBUG=y\n")
                    if not self.optimize:
                        f.write("CONFIG_DEBUG_OPTIMIZATIONS=y\n")
                else:
                    f.write("CONFIG_DEBUG=n\n")
                    if self.optimize:
                        if str(self.optimize) == "s":
                            f.write("CONFIG_SIZE_OPTIMIZATIONS=y\n")
                        elif str(self.optimize) == "0":
                            f.write("CONFIG_NO_OPTIMIZATIONS=y\n")
                        elif str(self.optimize) == "g":
                            f.write("CONFIG_DEBUG_OPTIMIZATIONS=y\n")
                        elif str(self.optimize) == "2":
                            f.write("CONFIG_SPEED_OPTIMIZATIONS=y\n")
                        elif str(self.optimize) == "3":
                            f.write("CONFIG_SPEED_OPTIMIZATIONS=y\n")
                            f.write('CONFIG_COMPILER_OPT="-O3"\n')
                        else:
                            raise RuntimeError(f"Unsupported optimization level for Zephyr platform: {self.optimize}")
                # f.write("CONFIG_PARTITION_TABLE_SINGLE_APP_LARGE=y\n")
                # if self.debug:
                #     f.write("CONFIG_OPTIMIZATION_LEVEL_DEBUG=y\n")
                #     f.write("CONFIG_COMPILER_OPTIMIZATION_LEVEL_DEBUG=y\n")
                # else:
                #     # Trying to reduce the binary size as much as possible
                #     # (https://docs.espressif.com/projects/esp-idf/en/latest/esp32/api-guides/performance/size.html)
                #     f.write("CONFIG_BOOTLOADER_LOG_LEVEL_NONE=y\n")
                #     f.write("CONFIG_BOOTLOADER_LOG_LEVEL=0\n")
                #     f.write("CONFIG_BOOT_ROM_LOG_ALWAYS_OFF=y\n")
                #     f.write("CONFIG_COMPILER_OPTIMIZATION_ASSERTIONS_DISABLE=y\n")
                #     f.write("CONFIG_COMPILER_OPTIMIZATION_ASSERTION_LEVEL=0\n")
                #     f.write("CONFIG_COMPILER_OPTIMIZATION_CHECKS_SILENT=y\n")
                #     f.write("CONFIG_HAL_DEFAULT_ASSERTION_LEVEL=0\n")
                #     f.write("CONFIG_LOG_DEFAULT_LEVEL_NONE=y\n")
                #     f.write("CONFIG_LOG_DEFAULT_LEVEL=0\n")
                #     f.write("CONFIG_LOG_MAXIMUM_LEVEL=0\n")
                #     f.write("CONFIG_NEWLIB_NANO_FORMAT=y\n")
                #     f.write("CONFIG_COMPILER_OPTIMIZATION_LEVEL_RELEASE=y\n")
                #     f.write("CONFIG_OPTIMIZATION_LEVEL_RELEASE=y\n")
                #     optimize_for_size = True
                #     if optimize_for_size:
                #         f.write("CONFIG_COMPILER_OPTIMIZATION_SIZE=y\n")
                #     else:
                #         f.write("CONFIG_COMPILER_OPTIMIZATION_PERF=y\n")
                # watchdog_sec = 60
                # f.write(f"CONFIG_ESP_TASK_WDT_TIMEOUT_S={watchdog_sec}\n")
                f.write(f'CONFIG_MLONMCU_CODEGEN_DIR="{src}"\n')
                for key, value in defs.items():
                    if isinstance(value, bool):
                        value = "y" if value else "n"
                    else:
                        value = f'"{value}"'
                    f.write(f"CONFIG_{key}={value}\n")

        write_defaults(self.project_dir / "prj.conf")
        zephyr_target = target.name.split("_", 1)[-1]
        westArgs = [  # cmake only
            "build",
            "-d",
            self.build_dir,
            "-b",
            zephyr_target,
            self.project_dir,
            *self.get_west_cmake_args(),
            "-c",
        ]
        return self.invoke_west(*westArgs, live=self.print_outputs)

    def compile(self, target, src=None):
        out = self.prepare(target, src)
        # TODO: build with cmake options
        # TODO: support self.num_threads
        # self.build_dir.mkdir()
        zephyr_target = target.name.split("_", 1)[-1]
        westArgs = [
            "build",
            "-d",
            self.build_dir,
            "-b",
            zephyr_target,
            self.project_dir,
            f"-o=-j{self.num_threads}",
        ]
        out += self.invoke_west(*westArgs, live=self.print_outputs)
        return out

    def generate(self, src, target, model=None) -> Tuple[dict, dict]:
        artifacts = []
        out = self.compile(target, src=src)
        elf_name = "zephyr.elf"
        elf_file = self.project_dir / "build" / "zephyr" / elf_name
        # TODO: just use path instead of raw data?
        if self.tempdir:
            # Warning: The porject context will get destroyed afterwards wehen using  a temporory directory
            with open(elf_file, "rb") as handle:
                data = handle.read()
                artifact = Artifact("generic_mlif", raw=data, fmt=ArtifactFormat.RAW)
                artifacts.append(artifact)
        else:
            artifact = Artifact(elf_name, path=elf_file, fmt=ArtifactFormat.PATH)
            artifacts.append(artifact)
        metrics = self.get_metrics(elf_file)
        stdout_artifact = Artifact(
            "zephyr_out.log", content=out, fmt=ArtifactFormat.TEXT  # TODO: split into one file per command
        )  # TODO: rename to tvmaot_out.log?
        artifacts.append(stdout_artifact)
        return {"default": artifacts}, {"default": metrics}

    def get_serial(self, target):
        port = target.port
        if port is None:
            # Falling back to 'espidf.baud'
            assert self.port is not None, f"Please provide '{target.name}.port'"
            port = self.port
        baud = target.baud
        if baud is None:
            # Falling back to 'espidf.baud'
            assert self.baud is not None, f"Please provide '{target.name}.baud'"
            baud = self.baud
        return port, baud

    def flash(self, elf, target, timeout=120):
        # Ignore elf, as we use self.project_dir instead
        # TODO: add alternative approach which allows passing elf instead
        if elf is not None:
            logger.debug("Ignoring ELF file for zephyr platform")
        # TODO: implement timeout
        # TODO: make sure that already compiled? -> error or just call compile routine?
        if self.wait_for_user:  # INTERACTIVE
            answer = input(
                f"Make sure that the device '{target.name}' is connected before you press [Enter]"
                + " (Type 'Abort' to cancel)"
            )
            if answer.lower() == "abort":
                return ""
        logger.debug("Flashing target software")

        westArgs = [
            "flash",
            "-d",
            self.build_dir,
        ]
        if "esp32" in target.name:  # TODO: implement this for other runners as well
            port, baud = self.get_serial(target)
            if port:
                westArgs.extend(["--esp-device", port])
            if baud:
                westArgs.extend(["--esp-baud-rate", baud])
        self.invoke_west(*westArgs, live=self.print_outputs)

    def monitor(self, target, timeout=60):
        if self.flash_only:
            return ""

        port, baud = self.get_serial(target)

        def _monitor_helper(port, baud, verbose=False, start_match=None, end_match=None, timeout=60):
            # Local import to get rid of pyserial dependency

            import serial

            # start_match and end_match are inclusive
            found_start = start_match is None
            outStr = ""
            if timeout:
                pass  # TODO: implement timeout
            # The following is a custom initialization sequence inspired by
            # (https://github.com/espressif/esp-idf/blob/master/tools/idf_monitor_base/serial_reader.py)
            # Required for esp32c3!
            high = False
            low = True
            ser = serial.Serial(port, baud)
            ser.close()
            ser.dtr = False
            ser.rts = False
            time.sleep(1)
            ser.dtr = low
            ser.rts = high
            ser.dtr = ser.dtr
            ser.open()
            ser.dtr = high
            ser.rts = low
            ser.dtr = ser.dtr
            time.sleep(0.002)
            ser.rts = high
            ser.dtr = ser.dtr
            try:
                while True:
                    try:
                        ser_bytes = ser.readline()
                        new_line = ser_bytes.decode("utf-8", errors="replace")
                        if verbose:
                            print(new_line.replace("\n", ""))
                        if start_match and start_match in new_line:
                            outStr = new_line
                            found_start = True
                        else:
                            outStr = outStr + new_line
                        if found_start:
                            if end_match and end_match in new_line:
                                break
                    except KeyboardInterrupt:
                        logger.warning("Stopped processing serial port (KeyboardInterrupt)")
            finally:
                ser.close()
            return outStr

        logger.debug("Monitoring target software")
        # TODO: implement timeout
        return _monitor_helper(
            port,
            baud,
            verbose=self.print_outputs,
            start_match="MLonMCU: START",
            end_match="MLonMCU: STOP",
            timeout=timeout,
        )



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# from mlonmcu.setup.utils import (
#     makeFlags,
#     makeDirName,
#     mkdirs,
#     clone,
#     apply,
#     make,
#     cmake,
#     download,
#     extract,
#     remove,
#     move,
#     copy,
#     is_populated,
#     download_and_extract,
# )


def test_setup_utils_makeFlags():
    pass


def test_setup_makeDirName():
    pass


# def test_setup_exec():
#     pass
#
#
# def test_setup_exec_getout():
#     pass
#
#
# def test_setup_python():
#     pass


def test_setup_mkdirs():
    pass


def test_setup_clone():
    pass


def test_setup_apply():
    pass


def test_setup_make():
    pass


def test_setup_cmake():
    pass


def test_setup_download():
    pass


def test_setup_extract():
    pass


def test_setup_remove():
    pass


def test_setup_move():
    pass


def test_setup_copy():
    pass


def test_setup_is_populated():
    pass


def test_setup_download_and_extract():
    pass


# def makeFlags(*args):
# def makeDirName(base: str, *args, flags: list = None) -> str:
# def exec(*args, **kwargs):
# def exec_getout(*args, live: bool = False, print_output: bool = True, handle_exit=None, prefix="", **kwargs) -> str:
# def python(*args, **kwargs):
# def mkdirs(path: Union[str, bytes, os.PathLike]):
# def clone( url: str, dest: Union[str, bytes, os.PathLike], branch: str = "",
#     recursive: bool = False, refresh: bool = False,):
# def apply( repo_dir: Path, patch_file: Path,):
# def make(*args, threads=multiprocessing.cpu_count(), use_ninja=False, cwd=None, verbose=False, **kwargs):
# def cmake(src, *args, debug=False, use_ninja=False, cwd=None, **kwargs):
# def download(url, dest, progress=False):
# def extract(archive, dest, progress=False):
# def remove(path):
# def move(src, dest):
# def copy(src, dest):
# def is_populated(path):
# def download_and_extract(url, archive, dest, progress=False):



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Helper functions used by MLonMCU targets"""

import argparse
from typing import List

from mlonmcu.cli.helper.parse import extract_feature_names, extract_config
from mlonmcu.feature.type import FeatureType
from mlonmcu.feature.features import get_available_features
from mlonmcu.logging import get_logger
from mlonmcu.setup.utils import execute as execute_new

logger = get_logger()


def execute(*args, **kwargs):
    """Redicrection of old mlonmcu.target.common.execute to new location
    mlonmcu.setup.utils.execute

    Parameters
    ----------
    args : list
        Arguments
    kwargs : dict
        Keyword Arguments
    """
    logger.warning("DEPRECATED: Please use mlonmcu.setup.utils.execute instead of mlonmcu.target.common.execute")
    return execute_new(*args, **kwargs)


def add_common_options(parser: argparse.ArgumentParser, target):
    """Add a set of common options to a command line parser.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The command line parser
    """
    target_group = parser.add_argument_group("target options")
    target_group.add_argument(
        "-f",
        "--feature",
        type=str,
        metavar="FEATURE",
        action="append",
        choices=target.FEATURES,  # TODO: get from selected target?
        help="Enabled features for target (Choices: %(choices)s)",
    )
    target_group.add_argument(
        "-c",
        "--config",
        metavar="KEY=VALUE",
        nargs="+",
        action="append",
        help="Custom target config as key-value pairs",
    )


def init_target_features(names, config):
    features = []
    for name in names:
        avail_features = get_available_features(feature_type=FeatureType.TARGET, feature_name=name)
        for feature_class in avail_features.values():
            features.append(feature_class(config=config))
    return features


def cli(target, args: List[str] = None):
    """Utility to handle the command line api for targets.

    Parameters
    ----------
    target : Target
        The target to be used.
    args : list
        Interface to pass arguments to the command line parser from test functions.
    """
    parser = argparse.ArgumentParser(description="ML on MCU Target")
    subparsers = parser.add_subparsers(dest="subcommand")
    exec_parser = subparsers.add_parser("exec", description="Run program with target")

    def _handle_execute(args):
        config, _ = extract_config(args)
        feature_names, _ = extract_feature_names(args)
        features = init_target_features(feature_names, config)
        target_inst = target(features=features, config=config)
        target_inst.exec(args.program, *args.extra_args, live=True)

    exec_parser.set_defaults(func=_handle_execute)
    add_common_options(exec_parser, target)
    exec_group = exec_parser.add_argument_group("Exec options")
    exec_group.add_argument("program", metavar="EXE", type=str, help="The program which should be executed")
    exec_group.add_argument("extra_args", metavar="ARG", nargs="*", help="Additional arguments")
    inspect_parser = subparsers.add_parser("inspect", description="Inspect program with target")

    def _handle_inspect(args):
        config, _ = extract_config(args)
        feature_names, _ = extract_feature_names(args)
        features = init_target_features(feature_names, config)
        target_inst = target(features=features, config=config)
        target_inst.inspect(args.program)

    inspect_parser.set_defaults(func=_handle_inspect)
    add_common_options(inspect_parser, target)
    inspect_group = inspect_parser.add_argument_group("Inspect options")
    inspect_group.add_argument("program", metavar="EXE", type=str, help="The program which should be inspected")
    if args:
        args = parser.parse_args(args)
    else:
        args = parser.parse_args()
    if hasattr(args, "func"):
        args.func(args)
    else:
        raise RuntimeError("Invalid command. For usage details use '--help'!")

    # Argparge
    # Env Vars
    # MLONMCU Context (can be disabled via --plain)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLIF Platform"""
import os
import tempfile
from typing import Tuple

from pathlib import Path

from mlonmcu.config import str2bool
from mlonmcu.setup import utils  # TODO: Move one level up?
from mlonmcu.timeout import exec_timeout
from mlonmcu.artifact import Artifact, ArtifactFormat
from mlonmcu.logging import get_logger
from mlonmcu.target import get_targets
from mlonmcu.target.target import Target
from mlonmcu.models.utils import get_data_source

from ..platform import CompilePlatform, TargetPlatform
from .mlif_target import get_mlif_platform_targets, create_mlif_platform_target

logger = get_logger()


class MlifPlatform(CompilePlatform, TargetPlatform):
    """Model Library Interface Platform class."""

    FEATURES = (
        CompilePlatform.FEATURES
        | TargetPlatform.FEATURES
        | {
            "validate",
            "muriscvnn",
            "cmsisnn",
            "muriscvnnbyoc",
            "cmsisnnbyoc",
            "vext",
            "pext",
            "arm_mvei",
            "arm_dsp",
            "auto_vectorize",
            "benchmark",
            "xpulp",
        }  # TODO: allow Feature-Features with automatic resolution of initialization order
    )

    DEFAULTS = {
        **CompilePlatform.DEFAULTS,
        **TargetPlatform.DEFAULTS,
        "template": "ml_interface",
        "ignore_data": True,
        "skip_check": False,
        "fail_on_error": False,  # Prefer to add acolum with validation results instead of raising a RuntimeError
        "model_support_dir": None,
        "toolchain": "gcc",
        "prebuild_lib_path": None,
        "optimize": None,  # values: 0,1,2,3,s
        "input_data_path": None,
        "output_data_path": None,
        "mem_only": False,
        "debug_symbols": False,
        "verbose_makefile": False,
        "lto": False,
        "slim_cpp": True,
        "garbage_collect": True,
        "fuse_ld": None,
        "strip_strings": False,
        "unroll_loops": None,
        "goal": "generic_mlonmcu",  # Use 'generic_mlif' for older version of MLIF
    }

    REQUIRED = {"mlif.src_dir"}
    OPTIONAL = {"llvm.install_dir", "srecord.install_dir"}

    def __init__(self, features=None, config=None):
        super().__init__(
            "mlif",
            features=features,
            config=config,
        )
        self.tempdir = None
        self.build_dir = None

    @property
    def goal(self):
        return self.config["goal"]

    def gen_data_artifact(self):
        in_paths = self.input_data_path
        if not isinstance(in_paths, list):
            in_paths = [in_paths]
        in_paths_new = []
        for in_path in in_paths:
            if in_path is None:
                continue
            if not isinstance(in_path, Path):
                in_path = Path(in_path)
            if in_path.is_file():
                raise NotImplementedError
            elif in_path.is_dir():
                in_paths_new.extend([f for f in in_path.iterdir() if f.is_file()])
            else:
                return None
        in_paths = in_paths_new
        out_paths = self.output_data_path
        if not isinstance(out_paths, list):
            out_paths = [out_paths]
        out_paths_new = []
        for out_path in out_paths:
            if out_path is None:
                continue
            if not isinstance(out_path, Path):
                out_path = Path(out_path)
            if out_path.is_file():
                raise NotImplementedError
            elif out_path.is_dir():
                out_paths_new.extend([f for f in out_path.iterdir() if f.is_file()])
            else:
                return None
        out_paths = out_paths_new
        if len(in_paths) == 0 or len(out_paths) == 0:
            return None
        data_src = get_data_source(in_paths, out_paths)
        return Artifact("data.c", content=data_src, fmt=ArtifactFormat.SOURCE)

    def init_directory(self, path=None, context=None):
        if self.build_dir is not None:
            self.build_dir.mkdir(exist_ok=True)
            logger.debug("Build directory already initialized")
            return
        dir_name = self.name
        if path is not None:
            self.build_dir = Path(path)
        elif self.config["build_dir"]:
            self.build_dir = Path(self.config["build_dir"])
        else:
            if context:
                assert "temp" in context.environment.paths
                self.build_dir = (
                    context.environment.paths["temp"].path / dir_name
                )  # TODO: Need to lock this for parallel builds
            else:
                logger.debug(
                    "Creating temporary directory because no context was available "
                    "and 'mlif.build_dir' was not supplied"
                )
                self.tempdir = tempfile.TemporaryDirectory()
                self.build_dir = Path(self.tempdir.name) / dir_name
                logger.info("Temporary build directory: %s", self.build_dir)
        self.build_dir.mkdir(exist_ok=True)

    def create_target(self, name):
        assert name in self.get_supported_targets(), f"{name} is not a valid MLIF target"
        targets = get_targets()
        if name in targets:
            base = targets[name]
        else:
            base = Target
        return create_mlif_platform_target(name, self, base=base)

    @property
    def mlif_dir(self):
        return Path(self.config["mlif.src_dir"])

    @property
    def llvm_dir(self):
        return self.config["llvm.install_dir"]

    @property
    def srecord_dir(self):
        return self.config["srecord.install_dir"]

    @property
    def template(self):
        return self.config["template"]

    @property
    def ignore_data(self):
        value = self.config["ignore_data"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def skip_check(self):
        value = self.config["skip_check"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def fail_on_error(self):
        value = self.config["fail_on_error"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def validate_outputs(self):
        return not self.ignore_data

    @property
    def toolchain(self):
        return str(self.config["toolchain"])

    @property
    def model_support_dir(self):
        return self.config["model_support_dir"]

    @property
    def prebuild_lib_dir(self):
        return self.config["prebuild_lib_dir"]

    @property
    def optimize(self):
        return self.config["optimize"]

    @property
    def input_data_path(self):
        return self.config["input_data_path"]

    @property
    def output_data_path(self):
        return self.config["output_data_path"]

    @property
    def mem_only(self):
        value = self.config["mem_only"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def debug_symbols(self):
        value = self.config["debug_symbols"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def verbose_makefile(self):
        value = self.config["verbose_makefile"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def lto(self):
        value = self.config["lto"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def slim_cpp(self):
        value = self.config["slim_cpp"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def garbage_collect(self):
        value = self.config["garbage_collect"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def fuse_ld(self):
        value = self.config["fuse_ld"]
        return value

    @property
    def strip_strings(self):
        value = self.config["strip_strings"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def unroll_loops(self):
        value = self.config["unroll_loops"]
        if value is None:
            return None
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def get_supported_targets(self):
        target_names = get_mlif_platform_targets()
        return target_names

    def close(self):
        if self.tempdir:
            self.tempdir.cleanup()

    def get_definitions(self):
        definitions = self.definitions
        definitions["TEMPLATE"] = self.template
        definitions["TOOLCHAIN"] = self.toolchain
        definitions["QUIET"] = self.mem_only
        definitions["SKIP_CHECK"] = self.skip_check
        if self.num_threads is not None:
            definitions["SUBPROJECT_THREADS"] = self.num_threads
        if self.toolchain == "llvm" and self.llvm_dir is None:
            raise RuntimeError("Missing config variable: llvm.install_dir")
        else:
            definitions["LLVM_DIR"] = self.llvm_dir
        if self.optimize is not None:
            definitions["OPTIMIZE"] = self.optimize
        if self.debug_symbols is not None:
            definitions["DEBUG_SYMBOLS"] = self.debug_symbols
        if self.verbose_makefile is not None:
            definitions["CMAKE_VERBOSE_MAKEFILE"] = self.verbose_makefile
        if self.lto is not None:
            definitions["ENABLE_LTO"] = self.lto
        if self.garbage_collect is not None:
            definitions["ENABLE_GC"] = self.garbage_collect
        if self.slim_cpp is not None:
            definitions["SLIM_CPP"] = self.slim_cpp
        if self.model_support_dir is not None:
            definitions["MODEL_SUPPORT_DIR"] = self.model_support_dir
        if self.fuse_ld is not None:
            definitions["FUSE_LD"] = self.fuse_ld
        if self.strip_strings is not None:
            definitions["STRIP_STRINGS"] = self.strip_strings
        if self.unroll_loops is not None:
            definitions["UNROLL_LOOPS"] = self.unroll_loops

        return definitions

    def get_cmake_args(self):
        cmakeArgs = []
        definitions = self.get_definitions()
        for key, value in definitions.items():
            if isinstance(value, bool):
                value = "ON" if value else "OFF"
            cmakeArgs.append(f"-D{key}={value}")
        return cmakeArgs

    def prepare(self):
        self.init_directory()

    def prepare_environment(self):
        env = os.environ.copy()
        if self.srecord_dir:
            path_old = env["PATH"]
            path_new = f"{self.srecord_dir}:{path_old}"
            env["PATH"] = path_new
        return env

    def configure(self, target, src, _model):
        del target
        if not isinstance(src, Path):
            src = Path(src)
        cmakeArgs = self.get_cmake_args()
        if src.is_file():
            src = src.parent  # TODO deal with directories or files?
        if src.is_dir():
            cmakeArgs.append("-DSRC_DIR=" + str(src))
        else:
            raise RuntimeError("Unable to find sources!")
        artifacts = []
        if self.ignore_data:
            cmakeArgs.append("-DDATA_SRC=")
        else:
            data_artifact = self.gen_data_artifact()
            if data_artifact:
                data_file = self.build_dir / data_artifact.name
                data_artifact.export(data_file)
                cmakeArgs.append("-DDATA_SRC=" + str(data_file))
                artifacts.append(data_artifact)
            else:
                logger.warning("No validation data provided for model.")
        utils.mkdirs(self.build_dir)
        env = self.prepare_environment()
        out = utils.cmake(
            self.mlif_dir,
            *cmakeArgs,
            cwd=self.build_dir,
            debug=self.debug,
            live=self.print_outputs,
            env=env,
        )
        return out, artifacts

    def compile(self, target, src=None, model=None, data_file=None):
        out = ""
        if src:
            configure_out, artifacts = self.configure(target, src, model)
            out += configure_out
        env = self.prepare_environment()
        out += utils.make(
            self.goal,
            cwd=self.build_dir,
            threads=self.num_threads,
            live=self.print_outputs,
            env=env,
        )
        return out, artifacts

    def generate(self, src, target, model=None) -> Tuple[dict, dict]:
        # TODO: fix timeouts
        if self.validate_outputs:
            # some strange bug?
            self.timeout_sec = 0
        else:
            self.timeout_sec = 90
        self.timeout_sec = 0
        if self.timeout_sec > 0:
            out, artifacts = exec_timeout(
                self.timeout_sec,
                self.compile,
                target,
                src=src,
                model=model,
            )
        else:
            out, artifacts = self.compile(target, src=src, model=model)
        elf_file = self.build_dir / "bin" / self.goal
        map_file = self.build_dir / "linker.map"  # TODO: optional
        hex_file = self.build_dir / "bin" / "generic_mlonmcu.hex"
        path_file = self.build_dir / "bin" / "generic_mlonmcu.path"  # TODO: move to dumps
        asmdump_file = self.build_dir / "dumps" / "generic_mlonmcu.dump"  # TODO: optional
        srcdump_file = self.build_dir / "dumps" / "generic_mlonmcu.srcdump"  # TODO: optional

        # TODO: just use path instead of raw data?
        with open(elf_file, "rb") as handle:
            data = handle.read()
            artifact = Artifact("generic_mlonmcu", raw=data, fmt=ArtifactFormat.RAW)
            artifacts.insert(0, artifact)  # First artifact should be the ELF
        # for cv32e40p
        if hex_file.is_file():
            with open(hex_file, "rb") as handle:
                data = handle.read()
                artifact = Artifact("generic_mlonmcu.hex", raw=data, fmt=ArtifactFormat.RAW)
                artifacts.insert(1, artifact)
        # only for vicuna
        if path_file.is_file():
            with open(path_file, "r") as handle:
                data = handle.read()
                artifact = Artifact("generic_mlonmcu.path", content=data, fmt=ArtifactFormat.TEXT)
                artifacts.insert(1, artifact)
        if map_file.is_file():
            with open(map_file, "r") as handle:
                data = handle.read()
                artifact = Artifact("generic_mlonmcu.map", content=data, fmt=ArtifactFormat.TEXT)
                artifacts.append(artifact)
        if asmdump_file.is_file():
            with open(asmdump_file, "r") as handle:
                data = handle.read()
                artifact = Artifact(
                    "generic_mlonmcu.dump", content=data, fmt=ArtifactFormat.TEXT, flags=(self.toolchain,)
                )
                artifacts.append(artifact)
        if srcdump_file.is_file():
            with open(srcdump_file, "r") as handle:
                data = handle.read()
                artifact = Artifact(
                    "generic_mlonmcu.srcdump", content=data, fmt=ArtifactFormat.TEXT, flags=(self.toolchain,)
                )
                artifacts.append(artifact)
        metrics = self.get_metrics(elf_file)
        stdout_artifact = Artifact(
            "mlif_out.log", content=out, fmt=ArtifactFormat.TEXT
        )  # TODO: rename to tvmaot_out.log?
        artifacts.append(stdout_artifact)
        return {"default": artifacts}, {"default": metrics}



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import pytest
import mock
import tempfile
import urllib.request
from pathlib import Path
from io import BytesIO
from zipfile import ZipFile

# import mlonmcu  # TODO: fix this bad?


@pytest.fixture()
def fake_config_home(tmp_path):
    config_home = tmp_path / "cfg"
    config_home.mkdir()
    mlonmcu_config_home = config_home / "mlonmcu"
    mlonmcu_config_home.mkdir()
    patcher = mock.patch.dict(os.environ, {"XDG_CONFIG_HOME": str(config_home)})
    patcher.start()
    yield mlonmcu_config_home
    patcher.stop()


@pytest.fixture()
def fake_environment_directory(tmp_path):
    cwd = tmp_path / "home"
    cwd.mkdir()
    yield cwd


@pytest.fixture()
def fake_working_directory(tmp_path, monkeypatch):
    cwd = tmp_path / "cwd"
    cwd.mkdir()
    monkeypatch.chdir(str(cwd))
    yield cwd


@pytest.fixture()
def fake_context():
    class FakeTaskCache:
        def __init__(self):
            self._vars = {}

    class FakeEnvironment:
        def __init__(self):
            self.paths = {}  # TODO: get rid of PathConfig if possible?
            self.flags = {}
            self.vars = {}

    class FakeContext:
        def __init__(self):
            self.environment = FakeEnvironment()
            self.cache = FakeTaskCache()

    context = FakeContext()
    yield context


@pytest.fixture()
def example_elf_file(request, tmp_path):
    name = request.param
    elf_path = tmp_path / name
    url = f"https://github.com/JonathanSalwan/binary-samples/raw/master/{name}"
    urllib.request.urlretrieve(url, elf_path)
    yield str(elf_path)


@pytest.fixture()
def user_context():
    from mlonmcu.context.context import MlonMcuContext

    try:
        with MlonMcuContext(deps_lock="read") as context:
            yield context
    except RuntimeError:
        pytest.skip("User Environment not found!")


@pytest.fixture(scope="session")
def models_dir():
    with tempfile.TemporaryDirectory() as tmp_path:
        url = "https://codeload.github.com/tum-ei-eda/mlonmcu-models/zip/refs/heads/main"
        resp = urllib.request.urlopen(url)
        with ZipFile(BytesIO(resp.read())) as zip_file:
            zip_file.extractall(tmp_path)
        models_path = Path(tmp_path) / "mlonmcu-models-main"
        yield models_path



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU RISC-V utilities"""
import re


def split_extensions(inp):
    inp = inp[4:]
    # special case for non std conform zve32x/zve64x
    matches = re.compile(r"(?:zve(?:32|64)x)|(?:[^xz_])|(?:x[^xz_]+)|(?:z[^xz_]+)").findall(inp)
    return set(matches)


def sort_extensions_canonical(extensions, lower=False, unpack=False):
    """Utility to get the canonical architecture name string."""

    # See: https://riscv.org/wp-content/uploads/2017/05/riscv-spec-v2.2.pdf#table.22.1
    ORDER = [
        "I",
        "M",
        "A",
        "F",
        "D",
        "G",
        "Q",
        "L",
        "C",
        "B",
        "J",
        "T",
        "P",
        "V",
        "Z",
        "XCVMAC",
        "XCVMEM",
        "XCVBI",
        "XCVALU",
        "XCVBITMANIP",
        "XCVSIMD",
        "XCVHWLP",
        "X",
        "S",
        "SX",
    ]  # What about Z* extensions?
    extensions_new = extensions.copy()

    # make upper
    extensions_new = [x.upper() for x in extensions_new]

    if unpack:
        # Convert G into IMAFD
        if "G" in extensions_new:
            extensions_new = [x for x in extensions_new if x != "G"] + ["I", "M", "A", "F", "D"]
        # Remove duplicates
        extensions_new = list(set(extensions_new))

    def _get_index(x):
        if x in ORDER:
            return ORDER.index(x)
        else:
            for i, o in enumerate(ORDER):
                if x.startswith(o):
                    return i
            return ORDER.index("X") - 0.5  # Insert unknown keys right before custom extensions

    extensions_new.sort(key=lambda x: _get_index(x))

    if lower:
        extensions_new = [x.lower() for x in extensions_new]

    return extensions_new


def join_extensions(exts, merge=True):
    sep = ""
    ret = ""
    if merge:
        if "i" in exts and "m" in exts and "a" in exts and "f" in exts and "d" in exts:
            exts = ["g"] + [e for e in exts if e not in "imafd"]
    for ext in exts:
        length = len(ext)
        if sep == "_":
            assert length > 1, "default extensions should come before any custom or sub-extensions"
        if length > 1 and ext not in ["xpulpv2", "xpulpv3", "xcorev"]:
            sep = "_"
        if ret != "":
            ret += sep
        ret += ext
    return ret


def update_extensions(
    exts,
    embedded=None,
    compressed=None,
    atomic=None,
    multiply=None,
    pext=None,
    pext_spec=None,
    vext=None,
    elen=None,
    embedded_vext=None,
    vlen=None,
    fpu=None,
    minimal=True,
    bext=None,
    bext_spec=None,
    bext_zba=None,
    bext_zbb=None,
    bext_zbc=None,
    bext_zbs=None,
):
    # ret = exts.copy()
    require = set()
    ignore_exts = ["zifencei", "zicsr"]
    for ext in exts:
        if ext == "g":
            fpu = "double"
            atomic = True
            multiply = True
        elif embedded is None and ext == "e":
            embedded = True
        elif multiply is None and ext == "m":
            multiply = True
        elif atomic is None and ext == "a":
            atomic = True
        elif compressed is None and ext == "c":
            compressed = True
        elif fpu is None and ext == "f":
            fpu = "single"
        elif (fpu is None or fpu == "single") and ext == "d":
            fpu = "double"
        elif vext is None and ext == "v":
            vext = True
        elif pext is None and ext == "p":
            pext = True
        elif vlen is None and "zvl" in ext and "sseg" not in ext:
            vlen_ = int(ext[3:-1])
            if vlen is None or vlen_ > vlen:
                vlen = vlen_
        elif embedded_vext is None and "zve" in ext:
            vext = True
            embedded_vext = True
            elen_ = int(ext[3:-1])
            if elen is None or elen_ > elen:
                elen = elen_
        # elif bext is None and ext in ["zba", "zbb", "zbc", "zbs"]:
        #     bext = True
        elif bext_zba is None and ext == "zba":
            bext_zba = True
        elif bext_zbb is None and ext == "zbb":
            bext_zbb = True
        elif bext_zbc is None and ext == "zbc":
            bext_zbc = True
        elif bext_zbs is None and ext == "zbs":
            bext_zbs = True
        elif ext in ignore_exts:
            pass
        else:
            require.add(ext)
    if embedded:
        require.add("e")
    else:
        require.add("i")
    if atomic:
        require.add("a")
    if multiply:
        require.add("m")
    if compressed:
        require.add("c")
    if fpu == "single":
        require.add("f")
    elif fpu == "double":
        require.add("d")
        require.add("f")
    if pext:
        require.add("p")
    if bext_zba:
        require.add("zba")
    if bext_zbb:
        require.add("zbb")
    if bext_zbc:
        require.add("zbc")
    if bext_zbs:
        require.add("zbs")
    if vext:
        if elen is None:
            elen = 32
        assert elen in [32, 64], f"Unsupported ELEN: {elen}"
        if elen == 32:  # Required to tell the compiler that EEW=64 is not allowed...
            if embedded_vext:
                if fpu in ["double", "single"]:
                    require.add("zve32f")
                else:
                    require.add("zve32x")
            else:
                assert fpu == "double"
                require.add("v")
        elif elen == 64:
            if embedded_vext:
                if fpu == "double":
                    require.add("zve64d")
                elif fpu == "single":
                    require.add("zve64f")
                else:
                    require.add("zve64x")
            else:
                assert fpu == "double"
                require.add("v")
        # if vlen:
        #     require.add(f"zvl{vlen}b")

    if not minimal:
        if fpu in ["single", "double"] and not minimal:
            require.add("zicsr")
        if vext or embedded_vext:
            require.add("zicsr")
        if atomic and multiply and fpu == "double":
            require.add("zifencei")

    ret = set()
    for ext in require:
        if ext not in ret:
            ret.add(ext)
    return ret


def update_extensions_pulp(exts, xpulp_version):
    ret = exts.copy()
    required = []
    if xpulp_version:
        required.append(f"xpulpv{xpulp_version}")
    for ext in required:
        if ext not in ret:
            ret.append(ext)
    return set(ret)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for the build process."""

from mlonmcu.flow import get_available_backend_names
from mlonmcu.cli.common import kickoff_runs
from mlonmcu.cli.load import handle as handle_load, add_load_options
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.session.run import RunStage
from mlonmcu.platform.lookup import get_platforms_targets, get_platforms_backends
from .helper.parse import (
    extract_backend_names,
    extract_target_names,
    extract_platform_names,
    extract_config_and_feature_names,
)


def add_build_options(parser):
    # TODO: rename to build_group
    add_load_options(parser)
    build_parser = parser.add_argument_group("build options")
    build_parser.add_argument(
        "-b",
        "--backend",
        type=str,
        action="append",
        choices=get_available_backend_names(),
        help="Backends to use (default: %(default)s)",
    )


def get_parser(subparsers, parent=None):
    """ "Define and return a subparser for the build subcommand."""
    parser = subparsers.add_parser(
        "build",
        description="Build model using the ML on MCU flow.",
        parents=[parent] if parent else [],
        add_help=(parent is None),
    )
    parser.set_defaults(flow_func=handle)
    add_build_options(parser)
    return parser


def _handle(args, context, require_target=False):
    handle_load(args, ctx=context)
    backends = extract_backend_names(args, context=context)
    targets = extract_target_names(args, context=context if require_target else None)
    platforms = extract_platform_names(args, context=context)

    new_config, _, _, _ = extract_config_and_feature_names(args, context=context)
    platform_backends = get_platforms_backends(context, config=new_config)  # This will be slow?
    platform_targets = get_platforms_targets(context, config=new_config)  # This will be slow?

    assert len(context.sessions) > 0
    session = context.sessions[-1]
    new_runs = []
    for run in session.runs:
        for target_name in targets:
            for backend_name in backends:
                new_run = run.copy()
                if backend_name is not None:
                    platform_name = None
                    for platform in platforms:
                        candidates = platform_backends[platform]
                        if backend_name in candidates:
                            try:
                                platform_name = platform
                                new_run.add_platform_by_name(platform_name, context=context)
                                break
                            except AssertionError:  # TODO: replace with incompatble error
                                platform_name = None
                                continue
                    # assert (
                    #     platform_name is not None
                    # ), f"Unable to find a suitable platform for the backend '{target_name}'"
                    # assert platform_name is not None

                if target_name is not None:
                    platform_name = None
                    for platform in platforms:
                        candidates = platform_targets[platform]
                        if target_name in candidates:
                            platform_name = platform
                    assert (
                        platform_name is not None
                    ), f"Unable to find a suitable platform for the target '{target_name}'"
                    new_run.add_platform_by_name(platform_name, context=context)
                    new_run.add_target_by_name(target_name, context=context)
                new_run.add_backend_by_name(backend_name, context=context)
                new_runs.append(new_run)

    session.runs = new_runs


def handle(args, ctx=None, require_target=False):
    if ctx:
        _handle(args, ctx, require_target=require_target)
    else:
        with MlonMcuContext(path=args.home, deps_lock="read") as context:
            _handle(args, context, require_target=require_target)
            kickoff_runs(args, RunStage.BUILD, context)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for invoking the mlonmcu flow."""

import sys

from mlonmcu.context.context import MlonMcuContext
import mlonmcu.platform.lookup
import mlonmcu.cli.load as load
import mlonmcu.cli.tune as tune
import mlonmcu.cli.build as build
import mlonmcu.cli.compile as compile_  # compile is a builtin name
import mlonmcu.cli.run as run
from mlonmcu.cli.common import add_flow_options, add_common_options, add_context_options


def get_parser(subparsers, parent=None):
    """ "Define and return a subparser for the flow subcommand."""
    parser = subparsers.add_parser(
        "flow",
        description="Invoke ML on MCU flow",
        parents=[parent] if parent else [],
        add_help=(parent is None),
    )
    parser.set_defaults(func=handle)
    add_common_options(parser)
    add_context_options(parser)
    add_flow_options(parser)
    subparsers = parser.add_subparsers(dest="subcommand2")  # this line changed
    load.get_parser(subparsers)
    tune.get_parser(subparsers)
    build.get_parser(subparsers)
    compile_.get_parser(subparsers)
    run.get_parser(subparsers)


def handle_list_targets(args):
    with MlonMcuContext(path=args.home, deps_lock="read") as context:
        mlonmcu.platform.lookup.print_summary(context=context)


def handle(args):
    """Callback function which will be called to process the flow subcommand"""
    if args.list_targets:
        handle_list_targets(args)
    else:
        if hasattr(args, "flow_func"):
            args.flow_func(args)
        else:
            print("Invalid command. Check 'mlonmcu flow --help' for the available subcommands!")
            sys.exit(1)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import os
import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_espidf(context: MlonMcuContext, params=None):
    return context.environment.has_platform("espidf")


@Tasks.provides(["espidf.src_dir"])
@Tasks.validate(_validate_espidf)
@Tasks.register(category=TaskType.PLATFORM)
def clone_espidf(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the ESP-IDF repository."""
    espidfName = utils.makeDirName("espidf")
    espidfSrcDir = context.environment.paths["deps"].path / "src" / espidfName
    user_vars = context.environment.vars
    if "espidf.src_dir" in user_vars:  # TODO: also check command line flags?
        return False
    if rebuild or not utils.is_populated(espidfSrcDir):
        espidfRepo = context.environment.repos["espidf"]
        utils.clone_wrapper(espidfRepo, espidfSrcDir, refresh=rebuild)
    context.cache["espidf.src_dir"] = espidfSrcDir


@Tasks.needs(["espidf.src_dir"])
@Tasks.provides(["espidf.install_dir"])
@Tasks.validate(_validate_espidf)
@Tasks.register(category=TaskType.PLATFORM)
def install_espidf(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Download and install target support for ESP-IDF toolchain."""
    espidfName = utils.makeDirName("espidf")
    espidfInstallDir = context.environment.paths["deps"].path / "install" / espidfName
    espidfSrcDir = context.cache["espidf.src_dir"]  # TODO: This will fail if the espidf.src_dir is user-supplied
    user_vars = context.environment.vars
    if "espidf.install_dir" in user_vars:  # TODO: also check command line flags?
        return False
    boards = ["all"]
    if "espidf.boards" in user_vars:
        boards = user_vars["espidf.boards"]
    if not isinstance(boards, str):
        assert isinstance(boards, list)
        boards = ",".join(boards)
    if not utils.is_populated(espidfInstallDir) or rebuild:
        # Using idf_tools.py directory instead of ./install.sh because we
        # don't want to use espe-idfs python environment
        espidfInstallScript = Path(espidfSrcDir) / "tools" / "idf_tools.py"
        espidfInstallArgs = ["install", f"--targets={boards}"]
        env = os.environ.copy()
        env["IDF_TOOLS_PATH"] = str(espidfInstallDir)
        utils.python(espidfInstallScript, *espidfInstallArgs, live=verbose, env=env)
    context.cache["espidf.install_dir"] = espidfInstallDir
    context.export_paths.add(espidfInstallDir / "bin")



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
from pathlib import Path
import venv
import os
from .list import get_environment_names, get_alternative_name, register_environment
from .config import (
    get_environments_dir,
    env_subdirs,
    get_config_dir,
    get_environments_file,
    DEFAULTS,
    init_config_dir,
)
from .templates import write_environment_yaml_from_template
from mlonmcu.utils import in_virtualenv, ask_user
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

logger = get_logger()


def create_environment_directories(path, directories):
    if not isinstance(path, Path):
        path = Path(path)
    if not path.is_dir():
        raise RuntimeError(f"Not a diretory: {path}")
    for directory in directories:
        (path / directory).mkdir(parents=True, exist_ok=True)


def clone_models_repo(dest, url=None, ref=None, refresh=False, recursive=False):
    if url is None:
        url = "https://github.com/tum-ei-eda/mlonmcu-models.git"
    utils.clone(url, dest, ref, refresh=refresh, recursive=False)


def create_venv_directory(base, hidden=True):
    if not isinstance(base, Path):
        base = Path(base)
    dirname = ".venv" if hidden else "venv"
    venv_dir = base / dirname
    venv.create(venv_dir)
    print(f"Virtual environment was created in {venv_dir}. Make sure to activate it before using mlonmcu.")


def initialize_environment(
    directory,
    name,
    interactive=True,
    create_venv=None,
    clone_models=None,
    allow_exists=None,
    register=None,
    template=None,
    config=None,
):
    overwrite = False
    assert template is not None
    print("Initializing ML on MCU environment")
    use_default_dir = directory == get_environments_dir()
    has_name = len(name.strip()) > 0
    if has_name:
        final_name = name.strip()
    else:
        if use_default_dir:
            final_name = DEFAULTS["environment"]
            has_name = True
        else:
            final_name = "unnamed"
    if use_default_dir:
        target_dir = os.path.join(directory, final_name)
    else:
        target_dir = directory
    target_dir = os.path.abspath(target_dir)
    config_dir = get_config_dir()
    if use_default_dir or register is not False:
        if not os.path.exists(config_dir):
            print(
                f"The mlonmcu user config directory {config_dir} does not exist!",
                end=" - ",
            )
            if not ask_user("Initialize?", default=True, interactive=interactive):
                print("Aborting...")
                sys.exit(1)
            init_config_dir()
            print("Initialized config directory.")

    print("Selected target directory:", target_dir)
    if os.path.exists(target_dir):
        print("The directory already exists!")
        if len(os.listdir(target_dir)) > 0:
            print("The directory is not empty!", end=" - ")
            # TODO: check for mlonmcu project files, if yes ask for overwrite instead
            if allow_exists is False or (
                allow_exists is None and not ask_user("Use anyway?", default=False, interactive=interactive)
            ):
                print("Aborting...")
                sys.exit(1)
        print("Using existing directory.")
        if has_name:  # This is a hack
            overwrite = True
    else:
        print("The directory does not exist!", end=" - ")
        if not ask_user("Create directory?", default=True, interactive=interactive):
            print("Aborting...")
            sys.exit(1)
        Path(target_dir).mkdir()
        print("Created directory.")
    print(f"Creating environment.yml based on template '{template}'.")
    # TODO: create and maintain environments.yml in user directory?
    write_environment_yaml_from_template(
        os.path.join(target_dir, "environment.yml"), template, home_dir=target_dir, config=config
    )

    # FIXME: controversial?
    if create_venv is None:
        if not in_virtualenv():
            print("It is strongly recommended to use mlonmcu inside a virtual Python environment.")
            if ask_user("Create one automatically?", default=False, interactive=interactive):
                # TODO: create venv
                create_venv_directory(target_dir)
        else:
            print("Skipping creation of virtual environment. (already inside one)")
    else:
        if create_venv:
            print("The creation of a virtual environment was requested")
            create_venv_directory(target_dir)
        else:
            print("Skipping creation of virtual environment.")

    subdirs = env_subdirs
    models_subdir = Path(target_dir) / "models"
    if not models_subdir.is_dir():
        if clone_models or (
            clone_models is None
            and ask_user(
                "Clone mlonmcu-models repository into environment?",
                default=True,
                interactive=interactive,
            )
        ):
            custom_url = None
            custom_ref = None
            custom_url = config.get("models.clone_url", None)
            custom_ref = config.get("models.clone_ref", None)
            clone_models_repo(models_subdir, url=custom_url, ref=custom_ref)
        else:
            subdirs.append("models")

    print("Initializing directories in environment:", " ".join(subdirs))
    create_environment_directories(target_dir, subdirs)

    if register or (
        register is None
        and ask_user(
            "Should the new environment be added to your list of environments?",
            default=has_name,
            interactive=interactive,
        )
    ):
        environments_file = get_environments_file()
        if not os.path.isfile(environments_file):
            print(f"Environments file ({environments_file}) does not exist!", end=" - ")
            if ask_user("Create empty one?", default=True, interactive=interactive):
                open(environments_file, "a").close()
                print("Initialized empty environments file.")

        env_names = get_environment_names()
        if final_name in env_names and not overwrite:
            alternative_name = get_alternative_name(final_name, env_names)
            print(f"An environment with the name '{final_name}' already exists. Using '{alternative_name}' instead")
            final_name = alternative_name
        # TODO: interactively ask the user
        print("Adding new environment to environments file.")
        register_environment(final_name, target_dir, overwrite=overwrite)

    print(f"Finished. Please add `export MLONMCU_HOME={target_dir}` to your shell configuration to use it anywhere")
    print("Recommended next step: `mlonmcu setup` inside your virtual environment.")
    print("Optional: `mlonmcu setup -g`")



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU ETISS/Pulpino Target definitions"""


from mlonmcu.logging import get_logger
from mlonmcu.target.common import cli
from .etiss import EtissTarget

logger = get_logger()


class EtissPulpinoTarget(EtissTarget):
    """Target using a Pulpino-like VP running in the ETISS simulator"""

    REQUIRED = EtissTarget.REQUIRED | {"etiss.src_dir"}

    def __init__(self, name="etiss_pulpino", features=None, config=None):
        super().__init__(name, features=features, config=config)

    def get_ini_bool_config(self):
        ret = super().get_ini_bool_config()
        ret["arch.enable_semihosting"] = False
        return ret

    def get_platform_defs(self, platform):
        assert platform == "mlif"
        ret = super().get_platform_defs(platform)
        ret["ETISS_DIR"] = self.etiss_dir
        del ret["MEM_ROM_ORIGIN"]
        del ret["MEM_ROM_LENGTH"]
        del ret["MEM_RAM_ORIGIN"]
        del ret["MEM_RAM_LENGTH"]
        ret["PULPINO_ROM_START"] = self.rom_start
        ret["PULPINO_ROM_SIZE"] = self.rom_size
        ret["PULPINO_RAM_START"] = self.ram_start
        ret["PULPINO_RAM_SIZE"] = self.ram_size
        return ret


if __name__ == "__main__":
    cli(target=EtissPulpinoTarget)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import signal
import sys
import multiprocessing
import subprocess

# import logging
import tarfile
import zipfile
import shutil
import tempfile
import urllib.request
from pathlib import Path
from typing import Union, List, Callable, Optional
from git import Repo
from tqdm import tqdm

from mlonmcu import logging
from mlonmcu.environment.config import RepoConfig

logger = logging.get_logger()


def makeFlags(*args):
    """Resolve tuple-like arguments to a list of string.

    Parameters
    ----------
    args
        List of tuples of the form: [(True, "foo"), (False, "bar")]

    Returns
    -------
    dirname : str
        The generated directory name

    Examples
    --------
    >>> makeFlags((True, "foo"), (False, "bar"))
    ["foo"]
    """
    return [name for check, name in args if check]


def makeDirName(base: str, *args, flags: list = None) -> str:
    """Creates a directory name based on configuration values.

    Using snake_case style.

    Parameters
    ----------
    base : str
        Prefix of the filename to be generated.
    args
        List of tuples of the form: [(True, "foo"), (False, "bar")]
    flags : list
        Optional list of additional flags to be added.

    Returns
    -------
    dirname : str
        The generated directory name

    Examples
    --------
    >>> makeDirName("base", (True, "foo"), (False, "bar"), flags=["flag"])
    "base_foo_flag"
    """
    names = [base] + makeFlags(*args)
    if flags:
        names = names + flags
    return "_".join(names)


def exec(*args, **kwargs):
    """Execute a process with the given args and using the given kwards as Popen arguments.

    Parameters
    ----------
    args
        The command to be executed.
    kwargs
        Parameters to be passed to subprocess
    """
    logger.warning("DEPRECATED: Please use utils.execute(..., ignore_output=True) instead of utils.exec(...)")
    # Original implementation
    # logger.debug("- Executing: " + str(args))
    # if "cwd" in kwargs:
    #     logger.debug("- CWD: " + str(kwargs["cwd"]))
    # subprocess.run([i for i in args], **kwargs, check=True)

    # Call new implementation
    _ = execute(*args, ignore_output=True, live=False, print_func=None, **kwargs)


def exec_getout(*args, live: bool = False, print_output: bool = False, handle_exit=None, prefix="", **kwargs) -> str:
    """Execute a process with the given args and using the given kwards as Popen arguments and return the output.

    Parameters
    ----------
    args
        The command to be executed.
    live : bool
        If the stdout should be updated in real time.
    print_output : bool
        Print the output at the end on non-live mode.

    Returns
    -------
    output
        The text printed to the command line.
    """
    logger.warning("DEPRECATED: Please use utils.execute(...) instead of utils.exec_getout(...)")
    # Original implementation:
    # logger.debug("- Executing: " + str(args))
    # outStr = ""
    # if live:
    #     process = subprocess.Popen([i for i in args], **kwargs, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    #     try:
    #         for line in process.stdout:
    #             new_line = prefix + line.decode(errors="replace")
    #             outStr = outStr + new_line
    #             print(new_line.replace("\n", ""))
    #         exit_code = None
    #         while exit_code is None:
    #             exit_code = process.poll()
    #         if handle_exit is not None:
    #             exit_code = handle_exit(exit_code)
    #         assert exit_code == 0, "The process returned an non-zero exit code {}! (CMD: `{}`)".format(
    #             exit_code, " ".join(list(map(str, args)))
    #         )
    #     except KeyboardInterrupt:
    #         logger.debug("Interrupted subprocess. Sending SIGINT signal...")
    #         pid = process.pid
    #         os.kill(pid, signal.SIGINT)

    # else:
    #     try:
    #         p = subprocess.Popen([i for i in args], **kwargs, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    #         outStr = p.communicate()[0].decode(errors="replace")
    #         exit_code = p.poll()
    #         # outStr = p.stdout.decode(errors="replace")
    #         if print_output:
    #             logger.debug(prefix + outStr)
    #         if handle_exit is not None:
    #             exit_code = handle_exit(exit_code)
    #         if exit_code != 0:
    #             logger.error(outStr)
    #         assert exit_code == 0, "The process returned an non-zero exit code {}! (CMD: `{}`)".format(
    #             exit_code, " ".join(list(map(str, args)))
    #         )
    #     except KeyboardInterrupt:
    #         logger.debug("Interrupted subprocess. Sending SIGINT signal...")
    #         pid = p.pid
    #         os.kill(pid, signal.SIGINT)
    #     except subprocess.CalledProcessError as e:
    #         outStr = e.output.decode(errors="replace")
    #         logger.error(outStr)
    #         raise e
    # return outStr
    return execute(
        *args,
        ignore_output=False,
        live=live,
        # print_func=print,
        handle_exit=handle_exit,
        err_func=logger.error,
        prefix=prefix,
        **kwargs,
    )


def execute(
    *args: List[str],
    ignore_output: bool = False,
    live: bool = False,
    print_func: Callable = print,
    handle_exit: Optional[Callable] = None,
    err_func: Callable = logger.error,
    prefix: str = "",
    **kwargs,
) -> str:
    """Wrapper for running a program in a subprocess.

    Parameters
    ----------
    args : list
        The actual command.
    ignore_output : bool
        Do not get the stdout and stderr or the subprocess.
    live : bool
        Print the output line by line instead of only at the end.
    print_func : Callable
        Function which should be used to print sysout messages.
    handle_exit: Callable
        Handler for exit code.
    err_func : Callable
        Function which should be used to print errors.
    kwargs: dict
        Arbitrary keyword arguments passed through to the subprocess.

    Returns
    -------
    out : str
        The command line output of the command
    """
    # TODO: catch keyboardinterrupt
    logger.debug("- Executing: %s", str(args))
    if "cwd" in kwargs:
        logger.debug("- CWD: %s", str(kwargs["cwd"]))
    if ignore_output:
        assert not live
        subprocess.run(args, **kwargs, check=True)
        return None

    out_str = ""
    if live:
        with subprocess.Popen(
            args,
            **kwargs,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        ) as process:
            try:
                for line in process.stdout:
                    new_line = prefix + line.decode(errors="replace")
                    out_str = out_str + new_line
                    print_func(new_line.replace("\n", ""))
                exit_code = None
                while exit_code is None:
                    exit_code = process.poll()
                if handle_exit is not None:
                    exit_code = handle_exit(exit_code, out=out_str)
                assert exit_code == 0, "The process returned an non-zero exit code {}! (CMD: `{}`)".format(
                    exit_code, " ".join(list(map(str, args)))
                )
            except KeyboardInterrupt:
                logger.debug("Interrupted subprocess. Sending SIGINT signal...")
                pid = process.pid
                os.kill(pid, signal.SIGINT)
    else:
        try:
            p = subprocess.Popen([i for i in args], **kwargs, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            out_str = p.communicate()[0].decode(errors="replace")
            out_str = prefix + out_str
            exit_code = p.poll()
            # print_func(out_str)
            if handle_exit is not None:
                exit_code = handle_exit(exit_code, out=out_str)
            if exit_code != 0:
                err_func(out_str)
            assert exit_code == 0, "The process returned an non-zero exit code {}! (CMD: `{}`)".format(
                exit_code, " ".join(list(map(str, args)))
            )
        except KeyboardInterrupt:
            logger.debug("Interrupted subprocess. Sending SIGINT signal...")
            pid = p.pid
            os.kill(pid, signal.SIGINT)
        except subprocess.CalledProcessError as e:
            out_str = e.output.decode(errors="replace")
            err_func(out_str)
            raise e

    return out_str


def python(*args, **kwargs):
    """Run a python script with the current interpreter."""
    return execute(sys.executable, *args, **kwargs)


# Makes sure all directories at the given path are created.
def mkdirs(path: Union[str, bytes, os.PathLike]):
    """Wrapper for os.makedirs which handels the special case where the path already exits."""
    if not os.path.exists(path):
        os.makedirs(path)


# Clones a git repository at given url into given dir and switches to the given branch.
def clone(
    url: str,
    dest: Union[str, bytes, os.PathLike],
    branch: str = "",
    submodules: list = [],
    recursive: bool = False,
    refresh: bool = False,
):
    """Helper function for cloning a repository.

    Parameters
    ----------
    url : str
        Clone URL of the repository.
    dest : Path
        Destination directory path.
    branch : str
        Optional branch name or commit reference/tag.
    submodules : list of strings
        Only affects when recursive is true. Submodules to be updated. If empty, all submodules will be updated.
    recursive : bool
        If the clone should be done recursively.
    refesh : bool
        Enables switching the url/branch if the repo already exists
    """
    mkdirs(dest)

    def update_submodules():
        if recursive:
            if submodules:
                for submodule in submodules:
                    assert isinstance(submodule, str), f"Submodules should be a list of str. {submodule} is not str."
                repo.git.submodule("update", "--init", "--recursive", "--", *submodules)
            else:
                repo.git.submodule("update", "--init", "--recursive")

    if is_populated(dest):
        if refresh:
            repo = Repo(dest)
            # TODO: backup old remote?
            repo.remotes.origin.set_url(url)
            repo.remotes.origin.fetch()
            repo.git.checkout(branch)
            repo.git.pull("origin", branch)  # This should also work for specific commits
            update_submodules()
    else:
        if branch:
            repo = Repo.clone_from(url, dest, recursive=recursive, no_checkout=True)
            repo.git.checkout(branch)
            update_submodules()
        else:
            Repo.clone_from(url, dest, recursive=recursive)


def clone_wrapper(cfg: RepoConfig, dest: Union[str, bytes, os.PathLike], refresh: bool = False):
    clone(cfg.url, dest, branch=cfg.ref, submodules=cfg.submodules, recursive=cfg.recursive, refresh=refresh)


def apply(
    repo_dir: Path,
    patch_file: Path,
):
    """Helper function for applying a patch to a repository.

    Parameters
    ----------
    repo_dir : Path
        Clone directory of repository.
    patch_file : Path
        Path to patch file.
    """

    repo = Repo(repo_dir)
    repo.git.clean("-xdf")  # Undo all changes
    repo.git.apply(patch_file)


def make(*args, threads=multiprocessing.cpu_count(), use_ninja=False, cwd=None, verbose=False, **kwargs):
    if cwd is None:
        raise RuntimeError("Please always pass a cwd to make()")
    if isinstance(cwd, Path):
        cwd = str(cwd.resolve())
    # TODO: make sure that ninja is installed?
    extraArgs = []
    tool = "ninja" if use_ninja else "make"
    extraArgs.append("-j" + str(threads))
    cmd = [tool] + extraArgs + list(args)
    return execute(*cmd, cwd=cwd, **kwargs)


def cmake(src, *args, debug=False, use_ninja=False, cwd=None, **kwargs):
    if cwd is None:
        raise RuntimeError("Please always pass a cwd to cmake()")
    if isinstance(cwd, Path):
        cwd = str(cwd.resolve())
    buildType = "Debug" if debug else "Release"
    extraArgs = []
    extraArgs.append("-DCMAKE_BUILD_TYPE=" + buildType)
    if use_ninja:
        extraArgs.append("-GNinja")
    cmd = ["cmake", str(src)] + extraArgs + list(args)
    return execute(*cmd, cwd=cwd, **kwargs)


# def move(a, b):  # TODO: make every utility compatible with Paths!
#     # This can not handle cross file-system renames!
#     if not isinstance(a, Path):
#         a = Path(a)
#     if not isinstance(b, Path):
#         b = Path(b)
#     a.replace(b)


def download(url, dest, progress=False):
    def hook(t):
        """Wraps tqdm instance."""
        last_b = [0]

        def update_to(b=1, bsize=1, tsize=None):
            """
            b  : int, optional
                Number of blocks transferred so far [default: 1].
            bsize  : int, optional
                Size of each block (in tqdm units) [default: 1].
            tsize  : int, optional
                Total size (in tqdm units). If [default: None] remains unchanged.
            """
            if tsize is not None:
                t.total = tsize
            t.update((b - last_b[0]) * bsize)
            last_b[0] = b

        return update_to

    if progress:
        with tqdm(unit="B", unit_scale=True, unit_divisor=1024, miniters=1, desc="Downloading File") as t:
            urllib.request.urlretrieve(url, dest, reporthook=hook(t))
    else:
        urllib.request.urlretrieve(url, dest)


def extract(archive, dest, progress=False):
    ext = Path(archive).suffix[1:]

    def handle(f):
        if progress:
            members = f.getmembers()
            for m in tqdm(iterable=members, total=len(members), desc="Extracting..."):
                f.extract(m, dest)
        else:
            f.extractall(dest)

    if ext == "zip":
        with zipfile.ZipFile(archive) as zip_file:
            handle(zip_file)
    elif ext in ["tar", "gz", "xz", "tgz", "bz2"]:
        with tarfile.open(archive) as tar_file:
            handle(tar_file)
    else:
        raise RuntimeError("Unable to detect the archive type")


def remove(path):
    os.remove(path)


def move(src, dest):
    shutil.move(src, dest)


def copy(src, dest):
    shutil.copy(src, dest)


def symlink(src, dest):
    os.symlink(src, dest)


def is_populated(path):
    if not isinstance(path, Path):
        path = Path(path)
    return path.is_dir() and os.listdir(path.resolve())


def download_and_extract(url, archive, dest, progress=False, force=True):
    if isinstance(dest, str):
        dest = Path(dest)
    assert isinstance(dest, Path)
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_archive = os.path.join(tmp_dir, archive)
        base_name = Path(archive).stem
        if ".tar" in base_name:
            base_name = Path(base_name).stem
        if url[-1] != "/":
            url += "/"
        download(url + archive, tmp_archive, progress=progress)
        extract(tmp_archive, tmp_dir, progress=progress)
        remove(os.path.join(tmp_dir, tmp_archive))
        mkdirs(dest.parent)
        if (Path(tmp_dir) / base_name).is_dir():  # Archive contains a subdirectory with the same name
            move(os.path.join(tmp_dir, base_name), dest)
        else:
            contents = list(Path(tmp_dir).glob("*"))
            if len(contents) == 1:
                tmp_dir_new = Path(tmp_dir) / contents[0]
                if tmp_dir_new.is_dir():  # Archive contains a single subdirectory with a different name
                    tmp_dir = tmp_dir_new
            if dest.is_dir():
                assert force, f"Set force=True to replace destination {dest}"
                shutil.rmtree(dest)
            move(tmp_dir, dest)


def patch(path, cwd=None):
    raise NotImplementedError



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory
from .cv32e40p import _validate_cv32e40p

logger = get_logger()

Tasks = get_task_factory()


def _validate_corevverif(context: MlonMcuContext, params=None):
    if _validate_cv32e40p(context, params=params):
        return True
    return False


@Tasks.provides(["corevverif.src_dir"])
@Tasks.validate(_validate_corevverif)
@Tasks.register(category=TaskType.TARGET)
def clone_corevverif(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the core-v-verif repo."""
    name = utils.makeDirName("corevverif")
    srcDir = context.environment.paths["deps"].path / "src" / name
    user_vars = context.environment.vars
    if "corevverif.src_dir" in user_vars:  # TODO: also check command line flags?
        return False
    if rebuild or not utils.is_populated(srcDir):
        repo = context.environment.repos["corevverif"]
        utils.clone_wrapper(repo, srcDir, refresh=rebuild)
    context.cache["corevverif.src_dir"] = srcDir



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import shutil
import multiprocessing
from tqdm import tqdm

from mlonmcu.logging import get_logger
from mlonmcu.feature.type import FeatureType
from mlonmcu.feature.features import get_matching_features
from mlonmcu.config import filter_config, str2bool
from .tasks import get_task_factory
from .task import TaskGraph
from mlonmcu.utils import ask_user

logger = get_logger()


class Setup:
    """MLonMCU dependency management interface."""

    FEATURES = set()

    DEFAULTS = {
        "print_outputs": False,
        "num_threads": None,
    }

    REQUIRED = set()
    OPTIONAL = set()

    def __init__(self, features=None, config=None, context=None, tasks_factory=None):
        if not tasks_factory:
            tasks_factory = get_task_factory()
        self.config = config if config else {}
        self.features = self.process_features(features)
        self.config = filter_config(self.config, "setup", self.DEFAULTS, self.OPTIONAL, self.REQUIRED)
        self.context = context
        self.tasks_factory = tasks_factory
        self.num_threads = int(
            self.config["num_threads"] if self.config["num_threads"] else multiprocessing.cpu_count()
        )

    @property
    def verbose(self):
        value = self.config["print_outputs"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def clean_cache(self, interactive=True):
        assert self.context is not None
        deps_dir = self.context.environment.lookup_path("deps").path
        cache_file = deps_dir / "cache.ini"
        if cache_file.is_file():
            print(f"The dependency cache file ({cache_file}) will be removed.")
            if ask_user("Are your sure?", default=not interactive, interactive=interactive):
                print(f"Removing {cache_file} ...")
                os.remove(cache_file)

    def clean_dependencies(self, interactive=True):
        assert self.context is not None
        self.clean_cache(interactive=interactive)
        deps_dir = self.context.environment.lookup_path("deps").path
        subdirs = ["src", "build", "install"]
        print(f"All dependencies will be removed from {deps_dir}.")
        if ask_user("Are your sure?", default=not interactive, interactive=interactive):
            for subdir in subdirs:
                full_path = deps_dir / subdir
                print(f"Removing contents of {full_path} ...")
                shutil.rmtree(full_path, ignore_errors=True)
                full_path.mkdir(exist_ok=True)

    def process_features(self, features):
        if features is None:
            return []
        features = get_matching_features(features, FeatureType.SETUP)
        for feature in features:
            # Not need to list features explicitly
            # assert (
            #     feature.name in self.FEATURES
            # ), f"Incompatible feature: {feature.name}"
            feature.used = True
            feature.add_setup_config(self.config)
        return features

    def _get_task_graph(self):
        self.tasks_factory.reset_changes()
        task_graph = TaskGraph(
            self.tasks_factory.registry.keys(),
            self.tasks_factory.dependencies,
            self.tasks_factory.providers,
        )
        return task_graph

    def get_dependency_order(self):
        task_graph = self._get_task_graph()
        V, E = task_graph.get_graph()
        order = task_graph.get_order()
        order_str = " -> ".join(order)
        logger.debug("Determined dependency order: %s" % order_str)
        return order

    def setup_progress_bar(self, enabled):
        if enabled:
            pbar = tqdm(
                total=len(self.tasks_factory.registry),
                desc="Installing dependencies",
                ncols=100,
                bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed}s]",
            )
            return pbar
        else:
            logger.info("Installing dependencies...")
            return None

    def write_cache_file(self):
        logger.debug("Updating dependency cache")
        cache_file = self.context.environment.paths["deps"].path / "cache.ini"
        self.context.cache.write_to_file(cache_file)

    def write_env_file(self):
        logger.debug("Updating paths cript")
        paths_file = self.context.environment.paths["deps"].path / "paths.sh"
        paths = self.context.export_paths
        temp = ":".join(map(str, paths))
        content = f"export PATH={temp}:$PATH"
        with open(paths_file, "w") as handle:
            handle.write(content)

    def visualize(self, path, ordered=False):
        task_graph = self._get_task_graph()
        task_graph.export_dot(path)
        logger.debug("Written task graph to file: %s" % path)

    def invoke_single_task(self, name, progress=False, write_cache=True, write_env=True, rebuild=False):
        assert name in self.tasks_factory.registry, f"Invalid task name: {name}"
        func = self.tasks_factory.registry[name]
        func(self.context, progress=progress, rebuild=rebuild, verbose=self.verbose, threads=self.num_threads)
        if write_cache:
            self.write_cache_file()
        if write_env:
            self.write_env_file()

    def install_dependencies(
        self,
        progress=False,
        write_cache=True,
        write_env=True,
        rebuild=False,
    ):
        assert self.context is not None
        order = self.get_dependency_order()
        pbar = self.setup_progress_bar(progress)
        for task in order:
            func = self.tasks_factory.registry[task]
            func(self.context, progress=progress, rebuild=rebuild, verbose=self.verbose, threads=self.num_threads)
            if pbar:
                pbar.update(1)
        if pbar:
            pbar.close()
        if write_cache:
            self.write_cache_file()
        if write_env:
            self.write_env_file()
        logger.info("Finished installing dependencies")
        return True

    def generate_requirements(
        self,
    ):
        from .gen_requirements import join_requirements
        import os

        requirements = join_requirements()
        output_dir = os.path.join(self.context.environment.home, "requirements_addition.txt")

        def feature_enabled_and_supported(obj, feature):
            if getattr(obj, "enabled", True):
                if getattr(obj, "name", None) == feature and getattr(obj, "supported", False):
                    return True
                else:
                    enable_and_supported = False
                    for k, v in obj.__dict__.items():
                        if isinstance(v, list):
                            for i in v:
                                enable_and_supported = enable_and_supported or feature_enabled_and_supported(i, feature)
                    return enable_and_supported
            else:
                return False

        with open(output_dir, "w") as f:
            f.write(f"# AUTOGENERATED by 'mlonmcu setup -g' {os.linesep}")
            config_pools = (
                self.context.environment.frameworks
                + self.context.environment.frontends
                + self.context.environment.platforms
                + self.context.environment.targets
            )
            for config in config_pools:
                if "espidf" in config.name and config.enabled:
                    for d in requirements["espidf"][1]:
                        f.write(f"{d}{os.linesep}")
                    logger.info("add dependencies for espidf")
                    break
            for config in config_pools:
                if "zephyr" in config.name and config.enabled:
                    for d in requirements["zephyr"][1]:
                        f.write(f"{d}{os.linesep}")
                    logger.info("add dependencies for zephyr")
                    break
            for config in config_pools:
                if "tflm" in config.name and config.enabled:
                    for d in requirements["tflm"][1]:
                        f.write(f"{d}{os.linesep}")
                    logger.info("add dependencies for tflm")
                    break
            for config in config_pools:
                if "tflite" in config.name and config.enabled:
                    for d in requirements["tflite"][1]:
                        f.write(f"{d}{os.linesep}")
                    logger.info("add dependencies for tflite")
                    break
            for config in config_pools:
                if "onnx" in config.name and config.enabled:
                    for d in requirements["onnx"][1]:
                        f.write(f"{d}{os.linesep}")
                    logger.info("add dependencies for onnx")
                    break
            for config in config_pools:
                if "etiss" in config.name and config.enabled:
                    for d in requirements["etiss"][1]:
                        f.write(f"{d}{os.linesep}")
                    logger.info("add dependencies for etiss")
                    break
            for config in config_pools:
                if "gvsoc_pulp" in config.name and config.enabled:
                    for d in requirements["gvsoc_pulp"][1]:
                        f.write(f"{d}{os.linesep}")
                    logger.info("add dependencies for gvsoc_pulp")
                    break
            for config in config_pools:
                if "visualize" in config.name and config.enabled:
                    for d in requirements["visualize"][1]:
                        f.write(f"{d}{os.linesep}")
                    logger.info("add dependencies for visualize")
                    break
            for config in config_pools:
                if "microtvm" in config.name and config.enabled:
                    for d in requirements["microtvm"][1]:
                        f.write(f"{d}{os.linesep}")
                    logger.info("add dependencies for microtvm")
                    break
            for config in config_pools:
                if feature_enabled_and_supported(config, "moiopt"):
                    for d in requirements["moiopt"][1]:
                        f.write(f"{d}{os.linesep}")
                    logger.info("add dependencies for moiopt")
                    break
            # for config in config_pools:
            # if feature_enabled_and_supported(config, "relayviz"):
            #       # equirements for relay visualization
            #     for d in requirements['relayviz'][1]:
            #         f.write(f"{d}{os.linesep}")
            # break
            for config in config_pools:
                if "tvm" in config.name and config.enabled:
                    for d in requirements["tvm"][1]:
                        f.write(f"{d}{os.linesep}")
                    logger.info("add dependencies for tvm")
                    break
            for config in config_pools:
                if "tvm" in config.name and feature_enabled_and_supported(config, "autotuned"):
                    for d in requirements["tvm-autotuning"][1]:
                        f.write(f"{d}{os.linesep}")
                    logger.info("add dependencies for tvm-autotuning")
                    break
            for config in config_pools:
                if "tf" in config.name and feature_enabled_and_supported(config, "visualize"):
                    # requirements for tvm tflite visualization
                    for d in requirements["visualize_tflite"][1]:
                        f.write(f"{d}{os.linesep}")
                    logger.info("add dependencies for tflite visualization")
                    break
        logger.info("Finished generating requirements_addition.txt")
        return True



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definitions of mlonmcu config templates."""
import pkgutil
import os
import jinja2
from pathlib import Path
import pkg_resources

from .config import get_config_dir


def get_template_names():
    template_files = pkg_resources.resource_listdir("mlonmcu", os.path.join("..", "resources", "templates"))
    names = [name.split(".yml.j2")[0] for name in template_files]
    return names


def get_template_text(name):
    return pkgutil.get_data("mlonmcu", os.path.join("..", "resources", "templates", name + ".yml.j2"))


def fill_template(name, data={}):
    if isinstance(name, Path):
        name = str(name)
    if name.endswith(".j2"):  # Template from file
        assert Path(name).is_file(), f"Template does not exits: {name}"
        with open(name, "r") as handle:
            template_text = handle.read()
    else:  # Template by name
        template_text = get_template_text(name)
    if template_text:
        if not isinstance(template_text, str):
            try:
                template_text = template_text.decode("utf-8")
            except UnicodeDecodeError as e:
                raise e
        tmpl = jinja2.Template(template_text)
        rendered = tmpl.render(**data)
        return rendered
    return None


def fill_environment_yaml(template_name, home_dir, config=None):
    if not config:
        config = {}
    return fill_template(template_name, {"home_dir": str(home_dir), "config_dir": str(get_config_dir()), **config})


def write_environment_yaml_from_template(path, template_name, home_dir, config=None):
    with open(path, "w") as yaml:
        text = fill_environment_yaml(template_name, home_dir, config=config)
        yaml.write(text)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Strategies for the q_vanilla_accelerator accelerator"""


from tvm import relay
from tvm.relay import op as _op
from tvm.relay.qnn.strategy.generic import wrap_topi_schedule, wrap_topi_qnn_conv2d
from tvm import topi


@relay.op.strategy.override_native_generic_func("qnn_conv2d_strategy")
def qnn_conv2d_strategy(attrs, inputs, out_type, target):
    print("qnn strategy")
    strategy = _op.OpStrategy()
    strategy.add_implementation(
        wrap_topi_qnn_conv2d(topi.hexagon.qnn_conv2d), wrap_topi_schedule(topi.hexagon.schedule_qnn_conv2d)
    )

    return strategy



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from mlonmcu.target.riscv.util import (
    sort_extensions_canonical,
    join_extensions,
    update_extensions,
    update_extensions_pulp,
)


def test_target_riscv_sort_extensions_canonical():
    # default
    assert sort_extensions_canonical(["c", "g"]) == ["G", "C"]
    assert sort_extensions_canonical(["i", "m", "a", "c", "f", "d"]) == ["I", "M", "A", "F", "D", "C"]
    assert sort_extensions_canonical(["g", "zifencei", "xcustom", "c"]) == ["G", "C", "ZIFENCEI", "XCUSTOM"]

    # lower
    assert sort_extensions_canonical(["c", "g"], lower=True) == ["g", "c"]

    # unpack
    assert sort_extensions_canonical(["c", "g"], unpack=True) == ["I", "M", "A", "F", "D", "C"]

    # lower + unpack
    assert sort_extensions_canonical(["c", "g"], lower=True, unpack=True) == ["i", "m", "a", "f", "d", "c"]

    # TODO: mixed cases


def test_target_riscv_join_extensions():
    assert join_extensions(["g", "c"]) == "gc"
    assert join_extensions(["g", "c", "zifencei"]) == "gc_zifencei"
    assert join_extensions(["g", "c", "xcustom"]) == "gc_xcustom"
    assert join_extensions(["g", "c", "xcorev"]) == "gcxcorev"

    with pytest.raises(AssertionError):
        join_extensions(["g", "xcustom", "c"])


def test_target_riscv_update_extensions():
    # default
    assert update_extensions(["i", "m", "c"]) == {"i", "m", "c"}

    # pext
    assert update_extensions(["i", "m", "c"], pext=True) == {"i", "m", "c", "p"}
    assert update_extensions(["i", "m", "c", "p"], pext=True) == {"i", "m", "c", "p"}
    # assert update_extensions(["i", "m", "c"], pext=True, variant="xuantie", pext_spec=0.94) == {
    #     "i",
    #     "m",
    #     "c",
    #     "p",
    #     "zpn",
    #     "zpsfoperand",
    # }
    # assert update_extensions(["i", "m", "c"], pext=True, variant="xuantie", pext_spec=0.97) == {
    #     "i",
    #     "m",
    #     "c",
    #     "p",
    #     "zpn",
    #     "zpsfoperand",
    #     "zbpbo",
    # }
    # assert update_extensions(["i", "m", "c", "p"], pext=False) == {"i", "m", "c", "?"}

    # vext
    assert update_extensions(["g", "c"], vext=True, fpu="double", elen=32) == {"i", "m", "a", "f", "d", "c", "v"}
    assert update_extensions(["g", "c"], vext=True, fpu="double", elen=64) == {"i", "m", "a", "f", "d", "c", "v"}
    # with pytest.raises(AssertionError):
    #     assert update_extensions(["g", "c"], vext=True, fpu="single")
    # with pytest.raises(AssertionError):
    #     assert update_extensions(["g", "c"], vext=True, fpu=None)

    # embedded vext
    base = {"i", "m", "a", "c"}
    assert update_extensions(base, vext=True, fpu=None, embedded_vext=True, elen=32) == {"i", "m", "a", "c", "zve32x"}
    assert update_extensions(base, vext=True, fpu=None, embedded_vext=True, elen=64) == {"i", "m", "a", "c", "zve64x"}
    assert update_extensions(base, vext=True, fpu="single", embedded_vext=True, elen=32) == {
        "i",
        "m",
        "a",
        "f",
        "c",
        "zve32f",
    }
    assert update_extensions(base, vext=True, fpu="single", embedded_vext=True, elen=64) == {
        "i",
        "m",
        "a",
        "f",
        "c",
        # "zve32f",
        "zve64f",
    }
    assert update_extensions(base, vext=True, fpu="double", embedded_vext=True, elen=64) == {
        "i",
        "m",
        "a",
        "f",
        "d",
        "c",
        # "zve32f",
        "zve64d",
    }

    # TODO: fpu, multiple elen


def test_target_riscv_update_extensions_pulp():
    assert update_extensions_pulp(["i", "m", "c"], None) == {"i", "m", "c"}
    assert update_extensions_pulp(["i", "m", "c"], 2) == {"i", "m", "c", "xpulpv2"}
